# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import datetime
from unittest.mock import Mock

from libcloud.test import MockHttp, unittest
from libcloud.utils.py3 import httplib, assertRaisesRegex
from libcloud.test.secrets import OPENSTACK_PARAMS
from libcloud.common.openstack import OpenStackBaseConnection
from libcloud.test.file_fixtures import ComputeFileFixtures
from libcloud.common.openstack_identity import (
    AUTH_TOKEN_EXPIRES_GRACE_SECONDS,
    OpenStackIdentityUser,
    OpenStackServiceCatalog,
    OpenStackIdentity_2_0_Connection,
    OpenStackIdentity_3_0_Connection,
    OpenStackIdentity_2_0_Connection_VOMS,
    OpenStackIdentity_3_0_Connection_AppCred,
    OpenStackIdentity_3_0_Connection_OIDC_access_token,
    get_class_for_auth_version,
)
from libcloud.compute.drivers.openstack import OpenStack_1_0_NodeDriver
from libcloud.test.compute.test_openstack import (
    OpenStackMockHttp,
    OpenStack_2_0_MockHttp,
    OpenStackMockAuthCache,
)

try:
    import simplejson as json
except ImportError:
    import json


TOMORROW = datetime.datetime.today() + datetime.timedelta(1)
YESTERDAY = datetime.datetime.today() - datetime.timedelta(1)


class OpenStackIdentityConnectionTestCase(unittest.TestCase):
    def setUp(self):
        OpenStackBaseConnection.auth_url = None
        OpenStackBaseConnection.conn_class = OpenStackMockHttp
        OpenStack_2_0_MockHttp.type = None
        OpenStackIdentity_3_0_MockHttp.type = None

    def test_auth_url_is_correctly_assembled(self):
        tuples = [
            ("1.0", OpenStackMockHttp, {}),
            ("1.1", OpenStackMockHttp, {}),
            ("2.0", OpenStack_2_0_MockHttp, {}),
            ("2.0_apikey", OpenStack_2_0_MockHttp, {}),
            ("2.0_password", OpenStack_2_0_MockHttp, {}),
            (
                "3.x_password",
                OpenStackIdentity_3_0_MockHttp,
                {"tenant_name": "tenant-name"},
            ),
            ("3.x_appcred", OpenStackIdentity_3_0_MockHttp, {}),
            (
                "3.x_oidc_access_token",
                OpenStackIdentity_3_0_MockHttp,
                {"tenant_name": "tenant-name"},
            ),
        ]

        auth_urls = [
            ("https://auth.api.example.com", ""),
            ("https://auth.api.example.com/", "/"),
            ("https://auth.api.example.com/foo/bar", "/foo/bar"),
            ("https://auth.api.example.com/foo/bar/", "/foo/bar/"),
        ]

        actions = {
            "1.0": "{url_path}/v1.0",
            "1.1": "{url_path}/v1.1/auth",
            "2.0": "{url_path}/v2.0/tokens",
            "2.0_apikey": "{url_path}/v2.0/tokens",
            "2.0_password": "{url_path}/v2.0/tokens",
            "3.x_password": "{url_path}/v3/auth/tokens",
            "3.x_appcred": "{url_path}/v3/auth/tokens",
            "3.x_oidc_access_token": "{url_path}/v3/OS-FEDERATION/identity_providers/user_name/protocols/tenant-name/auth",
        }

        user_id = OPENSTACK_PARAMS[0]
        key = OPENSTACK_PARAMS[1]

        for auth_version, mock_http_class, kwargs in tuples:
            for url, url_path in auth_urls:
                connection = self._get_mock_connection(
                    mock_http_class=mock_http_class, auth_url=url
                )

                auth_url = connection.auth_url
                cls = get_class_for_auth_version(auth_version=auth_version)
                osa = cls(
                    auth_url=auth_url,
                    user_id=user_id,
                    key=key,
                    parent_conn=connection,
                    **kwargs,
                )

                try:
                    osa = osa.authenticate()
                except Exception:
                    pass

                expected_path = actions[auth_version].format(url_path=url_path).replace("//", "/")

                self.assertEqual(
                    osa.action,
                    expected_path,
                )

    def test_basic_authentication(self):
        tuples = [
            ("1.0", OpenStackMockHttp, {}),
            ("1.1", OpenStackMockHttp, {}),
            ("2.0", OpenStack_2_0_MockHttp, {}),
            ("2.0_apikey", OpenStack_2_0_MockHttp, {}),
            ("2.0_password", OpenStack_2_0_MockHttp, {}),
            (
                "3.x_password",
                OpenStackIdentity_3_0_MockHttp,
                {
                    "user_id": "test_user_id",
                    "key": "test_key",
                    "token_scope": "project",
                    "tenant_name": "test_tenant",
                    "tenant_domain_id": "test_tenant_domain_id",
                    "domain_name": "test_domain",
                },
            ),
            (
                "3.x_appcred",
                OpenStackIdentity_3_0_MockHttp,
                {"user_id": "appcred_id", "key": "appcred_secret"},
            ),
            (
                "3.x_oidc_access_token",
                OpenStackIdentity_3_0_MockHttp,
                {
                    "user_id": "test_user_id",
                    "key": "test_key",
                    "token_scope": "domain",
                    "tenant_name": "test_tenant",
                    "tenant_domain_id": "test_tenant_domain_id",
                    "domain_name": "test_domain",
                },
            ),
        ]

        user_id = OPENSTACK_PARAMS[0]
        key = OPENSTACK_PARAMS[1]

        for auth_version, mock_http_class, kwargs in tuples:
            connection = self._get_mock_connection(mock_http_class=mock_http_class)
            auth_url = connection.auth_url

            if not kwargs:
                kwargs["user_id"] = user_id
                kwargs["key"] = key

            cls = get_class_for_auth_version(auth_version=auth_version)
            osa = cls(auth_url=auth_url, parent_conn=connection, **kwargs)

            self.assertEqual(osa.urls, {})
            self.assertIsNone(osa.auth_token)
            self.assertIsNone(osa.auth_user_info)
            osa = osa.authenticate()

            self.assertTrue(len(osa.urls) >= 1)
            self.assertTrue(osa.auth_token is not None)

            if auth_version in [
                "1.1",
                "2.0",
                "2.0_apikey",
                "2.0_password",
                "3.x_password",
                "3.x_appcred",
                "3.x_oidc_access_token",
            ]:
                self.assertTrue(osa.auth_token_expires is not None)

            if auth_version in [
                "2.0",
                "2.0_apikey",
                "2.0_password",
                "3.x_password",
                "3.x_appcred",
                "3.x_oidc_access_token",
            ]:
                self.assertTrue(osa.auth_user_info is not None)

    def test_token_expiration_and_force_reauthentication(self):
        user_id = OPENSTACK_PARAMS[0]
        key = OPENSTACK_PARAMS[1]

        connection = self._get_mock_connection(OpenStack_2_0_MockHttp)
        auth_url = connection.auth_url

        osa = OpenStackIdentity_2_0_Connection(
            auth_url=auth_url, user_id=user_id, key=key, parent_conn=connection
        )

        mocked_auth_method = Mock(wraps=osa._authenticate_2_0_with_body)
        osa._authenticate_2_0_with_body = mocked_auth_method

        # Force re-auth, expired token
        osa.auth_token = None
        osa.auth_token_expires = YESTERDAY
        count = 5

        for i in range(0, count):
            osa.authenticate(force=True)

        self.assertEqual(mocked_auth_method.call_count, count)

        # No force reauth, expired token
        osa.auth_token = None
        osa.auth_token_expires = YESTERDAY

        mocked_auth_method.call_count = 0
        self.assertEqual(mocked_auth_method.call_count, 0)

        for i in range(0, count):
            osa.authenticate(force=False)

        self.assertEqual(mocked_auth_method.call_count, 1)

        # No force reauth, valid / non-expired token
        osa.auth_token = None

        mocked_auth_method.call_count = 0
        self.assertEqual(mocked_auth_method.call_count, 0)

        for i in range(0, count):
            osa.authenticate(force=False)

            if i == 0:
                osa.auth_token_expires = TOMORROW

        self.assertEqual(mocked_auth_method.call_count, 1)

        # No force reauth, valid / non-expired token which is about to expire in
        # less than AUTH_TOKEN_EXPIRES_GRACE_SECONDS
        soon = datetime.datetime.utcnow() + datetime.timedelta(
            seconds=AUTH_TOKEN_EXPIRES_GRACE_SECONDS - 1
        )
        osa.auth_token = None

        mocked_auth_method.call_count = 0
        self.assertEqual(mocked_auth_method.call_count, 0)

        for i in range(0, count):
            if i == 0:
                osa.auth_token_expires = soon

            osa.authenticate(force=False)

        self.assertEqual(mocked_auth_method.call_count, 1)

    def test_authentication_cache(self):
        tuples = [
            # 1.0 does not provide token expiration, so it always
            # re-authenticates and never uses the cache.
            # ('1.0', OpenStackMockHttp, {}),
            ("1.1", OpenStackMockHttp, {}),
            ("2.0", OpenStack_2_0_MockHttp, {}),
            ("2.0_apikey", OpenStack_2_0_MockHttp, {}),
            ("2.0_password", OpenStack_2_0_MockHttp, {}),
            (
                "3.x_password",
                OpenStackIdentity_3_0_MockHttp,
                {
                    "user_id": "test_user_id",
                    "key": "test_key",
                    "token_scope": "project",
                    "tenant_name": "test_tenant",
                    "tenant_domain_id": "test_tenant_domain_id",
                    "domain_name": "test_domain",
                },
            ),
            (
                "3.x_oidc_access_token",
                OpenStackIdentity_3_0_MockHttp,
                {
                    "user_id": "test_user_id",
                    "key": "test_key",
                    "token_scope": "domain",
                    "tenant_name": "test_tenant",
                    "tenant_domain_id": "test_tenant_domain_id",
                    "domain_name": "test_domain",
                },
            ),
        ]

        user_id = OPENSTACK_PARAMS[0]
        key = OPENSTACK_PARAMS[1]

        for auth_version, mock_http_class, kwargs in tuples:
            mock_http_class.type = None
            connection = self._get_mock_connection(mock_http_class=mock_http_class)
            auth_url = connection.auth_url

            if not kwargs:
                kwargs["user_id"] = user_id
                kwargs["key"] = key

            auth_cache = OpenStackMockAuthCache()
            self.assertEqual(len(auth_cache), 0)
            kwargs["auth_cache"] = auth_cache

            cls = get_class_for_auth_version(auth_version=auth_version)
            osa = cls(auth_url=auth_url, parent_conn=connection, **kwargs)
            osa = osa.authenticate()

            # Token is cached
            self.assertEqual(len(auth_cache), 1)

            # New client, token from cache is re-used
            osa = cls(auth_url=auth_url, parent_conn=connection, **kwargs)
            osa.request = Mock(wraps=osa.request)
            osa = osa.authenticate()

            # No auth API call
            if auth_version in ("1.1", "2.0", "2.0_apikey", "2.0_password"):
                self.assertEqual(osa.request.call_count, 0)
            elif auth_version in ("3.x_password", "3.x_oidc_access_token"):
                # v3 only caches token and expiration; service catalog URLs
                # and the rest of the auth context are fetched from Keystone
                osa.request.assert_called_once_with(
                    action="/v3/auth/tokens",
                    params=None,
                    data=None,
                    headers={
                        "X-Subject-Token": "00000000000000000000000000000000",
                        "X-Auth-Token": "00000000000000000000000000000000",
                    },
                    method="GET",
                    raw=False,
                )

            # Cache size unchanged
            self.assertEqual(len(auth_cache), 1)

            # Authenticates if cached token expired
            cache_key = list(auth_cache.store.keys())[0]
            auth_context = auth_cache.get(cache_key)
            auth_context.expiration = YESTERDAY
            auth_cache.put(cache_key, auth_context)

            osa = cls(auth_url=auth_url, parent_conn=connection, **kwargs)
            osa.request = Mock(wraps=osa.request)
            osa._get_unscoped_token_from_oidc_token = Mock(return_value="000")
            OpenStackIdentity_3_0_MockHttp.type = "GET_UNAUTHORIZED_POST_OK"
            osa = osa.authenticate()

            if auth_version in ("1.1", "2.0", "2.0_apikey", "2.0_password"):
                self.assertEqual(osa.request.call_count, 1)
                self.assertTrue(osa.request.call_args[1]["method"], "POST")
            elif auth_version in ("3.x_password", "3.x_oidc_access_token"):
                self.assertTrue(osa.request.call_args[0][0], "/v3/auth/tokens")
                self.assertTrue(osa.request.call_args[1]["method"], "POST")

            # Token evicted from cache if 401 received on another call
            if hasattr(osa, "list_projects"):
                mock_http_class.type = None
                auth_cache.reset()

                osa = cls(auth_url=auth_url, parent_conn=connection, **kwargs)
                osa.request = Mock(wraps=osa.request)
                osa = osa.authenticate()
                self.assertEqual(len(auth_cache), 1)
                mock_http_class.type = "UNAUTHORIZED"
                try:
                    osa.list_projects()
                except:  # These methods don't handle 401s
                    pass
                self.assertEqual(len(auth_cache), 0)

    def _get_mock_connection(self, mock_http_class, auth_url=None):
        OpenStackBaseConnection.conn_class = mock_http_class

        if auth_url is None:
            auth_url = "https://auth.api.example.com"

        OpenStackBaseConnection.auth_url = auth_url
        connection = OpenStackBaseConnection(*OPENSTACK_PARAMS)

        connection._ex_force_base_url = "https://www.foo.com"
        connection.driver = OpenStack_1_0_NodeDriver(*OPENSTACK_PARAMS)

        return connection


class OpenStackIdentity_2_0_ConnectionTests(unittest.TestCase):
    def setUp(self):
        mock_cls = OpenStackIdentity_2_0_MockHttp
        mock_cls.type = None
        OpenStackIdentity_2_0_Connection.conn_class = mock_cls

        self.auth_instance = OpenStackIdentity_2_0_Connection(
            auth_url="http://none",
            user_id="test",
            key="test",
            tenant_name="test",
            proxy_url="http://proxy:8080",
            timeout=10,
        )
        self.auth_instance.auth_token = "mock"
        self.assertEqual(self.auth_instance.proxy_url, "http://proxy:8080")

    def test_list_projects(self):
        result = self.auth_instance.list_projects()
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0].id, "a")
        self.assertEqual(result[0].name, "test")
        self.assertEqual(result[0].description, "test project")
        self.assertTrue(result[0].enabled)


class OpenStackIdentity_3_0_ConnectionTests(unittest.TestCase):
    def setUp(self):
        mock_cls = OpenStackIdentity_3_0_MockHttp
        mock_cls.type = None
        OpenStackIdentity_3_0_Connection.conn_class = mock_cls

        self.auth_instance = OpenStackIdentity_3_0_Connection(
            auth_url="http://none",
            user_id="test",
            key="test",
            tenant_name="test",
            proxy_url="http://proxy:8080",
            timeout=10,
        )
        self.auth_instance.auth_token = "mock"
        self.assertEqual(self.auth_instance.proxy_url, "http://proxy:8080")

    def test_token_scope_argument(self):
        # Invalid token_scope value
        expected_msg = 'Invalid value for "token_scope" argument: foo'
        assertRaisesRegex(
            self,
            ValueError,
            expected_msg,
            OpenStackIdentity_3_0_Connection,
            auth_url="http://none",
            user_id="test",
            key="test",
            token_scope="foo",
        )

        # Missing tenant_name
        expected_msg = "Must provide tenant_name and domain_name argument"
        assertRaisesRegex(
            self,
            ValueError,
            expected_msg,
            OpenStackIdentity_3_0_Connection,
            auth_url="http://none",
            user_id="test",
            key="test",
            token_scope="project",
        )

        # Missing domain_name
        expected_msg = "Must provide domain_name argument"
        assertRaisesRegex(
            self,
            ValueError,
            expected_msg,
            OpenStackIdentity_3_0_Connection,
            auth_url="http://none",
            user_id="test",
            key="test",
            token_scope="domain",
            domain_name=None,
        )

        # Scope to project all ok
        OpenStackIdentity_3_0_Connection(
            auth_url="http://none",
            user_id="test",
            key="test",
            token_scope="project",
            tenant_name="test",
            domain_name="Default",
        )
        # Scope to domain
        OpenStackIdentity_3_0_Connection(
            auth_url="http://none",
            user_id="test",
            key="test",
            token_scope="domain",
            tenant_name=None,
            domain_name="Default",
        )

    def test_authenticate(self):
        auth = OpenStackIdentity_3_0_Connection(
            auth_url="http://none",
            user_id="test_user_id",
            key="test_key",
            token_scope="project",
            tenant_name="test_tenant",
            tenant_domain_id="test_tenant_domain_id",
            domain_name="test_domain",
            proxy_url="http://proxy:8080",
            timeout=10,
        )
        auth.authenticate()
        self.assertEqual(auth.proxy_url, "http://proxy:8080")

    def test_list_supported_versions(self):
        OpenStackIdentity_3_0_MockHttp.type = "v3"

        versions = self.auth_instance.list_supported_versions()
        self.assertEqual(len(versions), 2)
        self.assertEqual(versions[0].version, "v2.0")
        self.assertEqual(versions[0].url, "http://192.168.18.100:5000/v2.0/")
        self.assertEqual(versions[1].version, "v3.0")
        self.assertEqual(versions[1].url, "http://192.168.18.100:5000/v3/")

    def test_list_domains(self):
        domains = self.auth_instance.list_domains()
        self.assertEqual(len(domains), 1)
        self.assertEqual(domains[0].id, "default")
        self.assertEqual(domains[0].name, "Default")
        self.assertTrue(domains[0].enabled)

    def test_list_projects(self):
        projects = self.auth_instance.list_projects()
        self.assertEqual(len(projects), 4)
        self.assertEqual(projects[0].id, "a")
        self.assertEqual(projects[0].domain_id, "default")
        self.assertTrue(projects[0].enabled)
        self.assertEqual(projects[0].description, "Test project")

    def test_list_users(self):
        users = self.auth_instance.list_users()
        self.assertEqual(len(users), 12)
        self.assertEqual(users[0].id, "a")
        self.assertEqual(users[0].domain_id, "default")
        self.assertEqual(users[0].enabled, True)
        self.assertEqual(users[0].email, "openstack-test@localhost")

    def test_list_roles(self):
        roles = self.auth_instance.list_roles()
        self.assertEqual(len(roles), 2)
        self.assertEqual(roles[1].id, "b")
        self.assertEqual(roles[1].name, "admin")

    def test_list_user_projects(self):
        user = self.auth_instance.list_users()[0]
        projects = self.auth_instance.list_user_projects(user=user)
        self.assertEqual(len(projects), 0)

    def test_list_user_domain_roles(self):
        user = self.auth_instance.list_users()[0]
        domain = self.auth_instance.list_domains()[0]
        roles = self.auth_instance.list_user_domain_roles(domain=domain, user=user)
        self.assertEqual(len(roles), 1)
        self.assertEqual(roles[0].name, "admin")

    def test_get_domain(self):
        domain = self.auth_instance.get_domain(domain_id="default")
        self.assertEqual(domain.name, "Default")

    def test_get_user(self):
        user = self.auth_instance.get_user(user_id="a")
        self.assertEqual(user.id, "a")
        self.assertEqual(user.domain_id, "default")
        self.assertEqual(user.enabled, True)
        self.assertEqual(user.email, "openstack-test@localhost")

    def test_get_user_without_email(self):
        user = self.auth_instance.get_user(user_id="b")
        self.assertEqual(user.id, "b")
        self.assertEqual(user.name, "userwithoutemail")
        self.assertIsNone(user.email)

    def test_get_user_without_enabled(self):
        user = self.auth_instance.get_user(user_id="c")
        self.assertEqual(user.id, "c")
        self.assertEqual(user.name, "userwithoutenabled")
        self.assertIsNone(user.enabled)

    def test_create_user(self):
        user = self.auth_instance.create_user(
            email="test2@localhost", password="test1", name="test2", domain_id="default"
        )

        self.assertEqual(user.id, "c")
        self.assertEqual(user.name, "test2")

    def test_enable_user(self):
        user = self.auth_instance.list_users()[0]
        result = self.auth_instance.enable_user(user=user)
        self.assertTrue(isinstance(result, OpenStackIdentityUser))

    def test_disable_user(self):
        user = self.auth_instance.list_users()[0]
        result = self.auth_instance.disable_user(user=user)
        self.assertTrue(isinstance(result, OpenStackIdentityUser))

    def test_grant_domain_role_to_user(self):
        domain = self.auth_instance.list_domains()[0]
        role = self.auth_instance.list_roles()[0]
        user = self.auth_instance.list_users()[0]

        result = self.auth_instance.grant_domain_role_to_user(domain=domain, role=role, user=user)
        self.assertTrue(result)

    def test_revoke_domain_role_from_user(self):
        domain = self.auth_instance.list_domains()[0]
        role = self.auth_instance.list_roles()[0]
        user = self.auth_instance.list_users()[0]

        result = self.auth_instance.revoke_domain_role_from_user(
            domain=domain, role=role, user=user
        )
        self.assertTrue(result)

    def test_grant_project_role_to_user(self):
        project = self.auth_instance.list_projects()[0]
        role = self.auth_instance.list_roles()[0]
        user = self.auth_instance.list_users()[0]

        result = self.auth_instance.grant_project_role_to_user(
            project=project, role=role, user=user
        )
        self.assertTrue(result)

    def test_revoke_project_role_from_user(self):
        project = self.auth_instance.list_projects()[0]
        role = self.auth_instance.list_roles()[0]
        user = self.auth_instance.list_users()[0]

        result = self.auth_instance.revoke_project_role_from_user(
            project=project, role=role, user=user
        )
        self.assertTrue(result)


class OpenStackIdentity_3_0_Connection_AppCredTests(unittest.TestCase):
    def setUp(self):
        mock_cls = OpenStackIdentity_3_0_AppCred_MockHttp
        mock_cls.type = None
        OpenStackIdentity_3_0_Connection_AppCred.conn_class = mock_cls

        self.auth_instance = OpenStackIdentity_3_0_Connection_AppCred(
            auth_url="http://none",
            user_id="appcred_id",
            key="appcred_secret",
            proxy_url="http://proxy:8080",
            timeout=10,
        )
        self.auth_instance.auth_token = "mock"

    def test_authenticate(self):
        auth = OpenStackIdentity_3_0_Connection_AppCred(
            auth_url="http://none",
            user_id="appcred_id",
            key="appcred_secret",
            proxy_url="http://proxy:8080",
            timeout=10,
        )
        auth.authenticate()


class OpenStackIdentity_3_0_Connection_OIDC_access_token_federation_projectsTests(
    unittest.TestCase
):
    def setUp(self):
        mock_cls = OpenStackIdentity_3_0_federation_projects_MockHttp
        mock_cls.type = None
        OpenStackIdentity_3_0_Connection_OIDC_access_token.conn_class = mock_cls

        self.auth_instance = OpenStackIdentity_3_0_Connection_OIDC_access_token(
            auth_url="http://none",
            user_id="idp",
            key="token",
            tenant_name="oidc",
            proxy_url="http://proxy:8080",
            timeout=10,
        )
        self.auth_instance.auth_token = "mock"

    def test_authenticate(self):
        auth = OpenStackIdentity_3_0_Connection_OIDC_access_token(
            auth_url="http://none",
            user_id="idp",
            key="token",
            token_scope="project",
            tenant_name="oidc",
            proxy_url="http://proxy:8080",
            timeout=10,
        )
        auth.authenticate()


class OpenStackIdentity_3_0_Connection_OIDC_access_tokenTests(unittest.TestCase):
    def setUp(self):
        mock_cls = OpenStackIdentity_3_0_MockHttp
        mock_cls.type = None
        OpenStackIdentity_3_0_Connection_OIDC_access_token.conn_class = mock_cls

        self.auth_instance = OpenStackIdentity_3_0_Connection_OIDC_access_token(
            auth_url="http://none",
            user_id="idp",
            key="token",
            tenant_name="oidc",
            domain_name="project_name2",
        )
        self.auth_instance.auth_token = "mock"

    def test_authenticate(self):
        auth = OpenStackIdentity_3_0_Connection_OIDC_access_token(
            auth_url="http://none",
            user_id="idp",
            key="token",
            token_scope="project",
            tenant_name="oidc",
            domain_name="project_name2",
        )
        auth.authenticate()


class OpenStackIdentity_3_0_Connection_OIDC_access_token_project_idTests(unittest.TestCase):
    def setUp(self):
        mock_cls = OpenStackIdentity_3_0_MockHttp
        mock_cls.type = None
        OpenStackIdentity_3_0_Connection_OIDC_access_token.conn_class = mock_cls

        self.auth_instance = OpenStackIdentity_3_0_Connection_OIDC_access_token(
            auth_url="http://none",
            user_id="idp",
            key="token",
            tenant_name="oidc",
            domain_name="project_id2",
        )
        self.auth_instance.auth_token = "mock"

    def test_authenticate_valid_project_id(self):
        auth = OpenStackIdentity_3_0_Connection_OIDC_access_token(
            auth_url="http://none",
            user_id="idp",
            key="token",
            token_scope="project",
            tenant_name="oidc",
            domain_name="project_id2",
        )
        auth.authenticate()

    def test_authenticate_invalid_project_id(self):
        auth = OpenStackIdentity_3_0_Connection_OIDC_access_token(
            auth_url="http://none",
            user_id="idp",
            key="token",
            token_scope="project",
            tenant_name="oidc",
            domain_name="project_id100",
        )

        expected_msg = "Project project_id100 not found"
        self.assertRaisesRegex(ValueError, expected_msg, auth.authenticate)


class OpenStackIdentity_2_0_Connection_VOMSTests(unittest.TestCase):
    def setUp(self):
        mock_cls = OpenStackIdentity_2_0_Connection_VOMSMockHttp
        mock_cls.type = None
        OpenStackIdentity_2_0_Connection_VOMS.conn_class = mock_cls

        self.auth_instance = OpenStackIdentity_2_0_Connection_VOMS(
            auth_url="http://none", user_id=None, key="/tmp/proxy.pem", tenant_name="VO"
        )
        self.auth_instance.auth_token = "mock"

    def test_authenticate(self):
        auth = OpenStackIdentity_2_0_Connection_VOMS(
            auth_url="http://none",
            user_id=None,
            key="/tmp/proxy.pem",
            token_scope="test",
            tenant_name="VO",
        )
        auth.authenticate()


class OpenStackServiceCatalogTestCase(unittest.TestCase):
    fixtures = ComputeFileFixtures("openstack")

    def test_parsing_auth_v1_1(self):
        data = self.fixtures.load("_v1_1__auth.json")
        data = json.loads(data)
        service_catalog = data["auth"]["serviceCatalog"]

        catalog = OpenStackServiceCatalog(service_catalog=service_catalog, auth_version="1.0")
        entries = catalog.get_entries()
        self.assertEqual(len(entries), 3)

        entry = [e for e in entries if e.service_type == "cloudFilesCDN"][0]
        self.assertEqual(entry.service_type, "cloudFilesCDN")
        self.assertIsNone(entry.service_name)
        self.assertEqual(len(entry.endpoints), 2)
        self.assertEqual(entry.endpoints[0].region, "ORD")
        self.assertEqual(entry.endpoints[0].url, "https://cdn2.clouddrive.com/v1/MossoCloudFS")
        self.assertEqual(entry.endpoints[0].endpoint_type, "external")
        self.assertEqual(entry.endpoints[1].region, "LON")
        self.assertEqual(entry.endpoints[1].endpoint_type, "external")

    def test_parsing_auth_v2(self):
        data = self.fixtures.load("_v2_0__auth.json")
        data = json.loads(data)
        service_catalog = data["access"]["serviceCatalog"]

        catalog = OpenStackServiceCatalog(service_catalog=service_catalog, auth_version="2.0")
        entries = catalog.get_entries()
        self.assertEqual(len(entries), 10)

        entry = [e for e in entries if e.service_name == "cloudServers"][0]
        self.assertEqual(entry.service_type, "compute")
        self.assertEqual(entry.service_name, "cloudServers")
        self.assertEqual(len(entry.endpoints), 1)
        self.assertIsNone(entry.endpoints[0].region)
        self.assertEqual(entry.endpoints[0].url, "https://servers.api.rackspacecloud.com/v1.0/1337")
        self.assertEqual(entry.endpoints[0].endpoint_type, "external")

    def test_parsing_auth_v3(self):
        data = self.fixtures.load("_v3__auth.json")
        data = json.loads(data)
        service_catalog = data["token"]["catalog"]

        catalog = OpenStackServiceCatalog(service_catalog=service_catalog, auth_version="3.x")
        entries = catalog.get_entries()
        self.assertEqual(len(entries), 6)
        entry = [e for e in entries if e.service_type == "volume"][0]
        self.assertEqual(entry.service_type, "volume")
        self.assertIsNone(entry.service_name)
        self.assertEqual(len(entry.endpoints), 3)
        self.assertEqual(entry.endpoints[0].region, "regionOne")
        self.assertEqual(entry.endpoints[0].endpoint_type, "external")
        self.assertEqual(entry.endpoints[1].region, "regionOne")
        self.assertEqual(entry.endpoints[1].endpoint_type, "admin")
        self.assertEqual(entry.endpoints[2].region, "regionOne")
        self.assertEqual(entry.endpoints[2].endpoint_type, "internal")

    def test_get_public_urls(self):
        data = self.fixtures.load("_v2_0__auth.json")
        data = json.loads(data)
        service_catalog = data["access"]["serviceCatalog"]

        catalog = OpenStackServiceCatalog(service_catalog=service_catalog, auth_version="2.0")

        public_urls = catalog.get_public_urls(service_type="object-store")
        expected_urls = [
            "https://storage101.lon1.clouddrive.com/v1/MossoCloudFS_11111-111111111-1111111111-1111111",
            "https://storage101.ord1.clouddrive.com/v1/MossoCloudFS_11111-111111111-1111111111-1111111",
        ]
        self.assertEqual(public_urls, expected_urls)

    def test_get_regions(self):
        data = self.fixtures.load("_v2_0__auth.json")
        data = json.loads(data)
        service_catalog = data["access"]["serviceCatalog"]

        catalog = OpenStackServiceCatalog(service_catalog=service_catalog, auth_version="2.0")

        regions = catalog.get_regions(service_type="object-store")
        self.assertEqual(regions, ["LON", "ORD"])

        regions = catalog.get_regions(service_type="invalid")
        self.assertEqual(regions, [])

    def test_get_service_types(self):
        data = self.fixtures.load("_v2_0__auth.json")
        data = json.loads(data)
        service_catalog = data["access"]["serviceCatalog"]

        catalog = OpenStackServiceCatalog(service_catalog=service_catalog, auth_version="2.0")
        service_types = catalog.get_service_types()
        self.assertEqual(
            service_types,
            [
                "compute",
                "image",
                "network",
                "object-store",
                "rax:object-cdn",
                "volumev2",
                "volumev3",
            ],
        )

        service_types = catalog.get_service_types(region="ORD")
        self.assertEqual(service_types, ["rax:object-cdn"])

    def test_get_service_names(self):
        data = self.fixtures.load("_v2_0__auth.json")
        data = json.loads(data)
        service_catalog = data["access"]["serviceCatalog"]

        catalog = OpenStackServiceCatalog(service_catalog=service_catalog, auth_version="2.0")

        service_names = catalog.get_service_names()
        self.assertEqual(
            service_names,
            [
                "cinderv2",
                "cinderv3",
                "cloudFiles",
                "cloudFilesCDN",
                "cloudServers",
                "cloudServersOpenStack",
                "cloudServersPreprod",
                "glance",
                "neutron",
                "nova",
            ],
        )

        service_names = catalog.get_service_names(service_type="compute")
        self.assertEqual(
            service_names,
            ["cloudServers", "cloudServersOpenStack", "cloudServersPreprod", "nova"],
        )


class OpenStackIdentity_2_0_MockHttp(MockHttp):
    fixtures = ComputeFileFixtures("openstack_identity/v2")
    json_content_headers = {"content-type": "application/json; charset=UTF-8"}

    def _v2_0_tenants(self, method, url, body, headers):
        if method == "GET":
            body = self.fixtures.load("v2_0_tenants.json")
            return (
                httplib.OK,
                body,
                self.json_content_headers,
                httplib.responses[httplib.OK],
            )
        raise NotImplementedError()


class OpenStackIdentity_3_0_MockHttp(MockHttp):
    fixtures = ComputeFileFixtures("openstack_identity/v3")
    json_content_headers = {"content-type": "application/json; charset=UTF-8"}

    def _v3(self, method, url, body, headers):
        if method == "GET":
            body = self.fixtures.load("v3_versions.json")
            return (
                httplib.OK,
                body,
                self.json_content_headers,
                httplib.responses[httplib.OK],
            )
        raise NotImplementedError()

    def _v3_domains(self, method, url, body, headers):
        if method == "GET":
            body = self.fixtures.load("v3_domains.json")
            return (
                httplib.OK,
                body,
                self.json_content_headers,
                httplib.responses[httplib.OK],
            )
        raise NotImplementedError()

    def _v3_projects(self, method, url, body, headers):
        if method == "GET":
            body = self.fixtures.load("v3_projects.json")
            return (
                httplib.OK,
                body,
                self.json_content_headers,
                httplib.responses[httplib.OK],
            )
        raise NotImplementedError()

    def _v3_projects_UNAUTHORIZED(self, method, url, body, headers):
        if method == "GET":
            body = ComputeFileFixtures("openstack").load("_v3__auth.json")
            return (
                httplib.UNAUTHORIZED,
                body,
                self.json_content_headers,
                httplib.responses[httplib.UNAUTHORIZED],
            )
        raise NotImplementedError()

    def _v3_OS_FEDERATION_identity_providers_test_user_id_protocols_test_tenant_auth(
        self, method, url, body, headers
    ):
        if method == "GET":
            if "Authorization" not in headers:
                return (
                    httplib.UNAUTHORIZED,
                    "",
                    headers,
                    httplib.responses[httplib.OK],
                )

            if headers["Authorization"] == "Bearer test_key":
                response_body = ComputeFileFixtures("openstack").load("_v3__auth.json")
                response_headers = {
                    "Content-Type": "application/json",
                    "x-subject-token": "foo-bar",
                }
                return (
                    httplib.OK,
                    response_body,
                    response_headers,
                    httplib.responses[httplib.OK],
                )

            return (httplib.UNAUTHORIZED, "{}", headers, httplib.responses[httplib.OK])
        raise NotImplementedError()

    def _v3_auth_tokens(self, method, url, body, headers):
        if method == "GET":
            body = json.loads(ComputeFileFixtures("openstack").load("_v3__auth.json"))
            body["token"]["expires_at"] = TOMORROW.isoformat()
            headers = self.json_content_headers.copy()
            headers["x-subject-token"] = "00000000000000000000000000000000"
            return (
                httplib.OK,
                json.dumps(body),
                headers,
                httplib.responses[httplib.OK],
            )
        if method == "POST":
            status = httplib.OK
            data = json.loads(body)
            if "password" in data["auth"]["identity"]:
                if (
                    data["auth"]["identity"]["password"]["user"]["domain"]["name"] != "test_domain"
                    or data["auth"]["scope"]["project"]["domain"]["id"] != "test_tenant_domain_id"
                ):
                    status = httplib.UNAUTHORIZED

            body = ComputeFileFixtures("openstack").load("_v3__auth.json")
            headers = self.json_content_headers.copy()
            headers["x-subject-token"] = "00000000000000000000000000000000"
            return (status, body, headers, httplib.responses[httplib.OK])
        raise NotImplementedError()

    def _v3_auth_tokens_GET_UNAUTHORIZED_POST_OK(self, method, url, body, headers):
        if method == "GET":
            body = ComputeFileFixtures("openstack").load("_v3__auth_unauthorized.json")
            return (
                httplib.UNAUTHORIZED,
                body,
                self.json_content_headers,
                httplib.responses[httplib.UNAUTHORIZED],
            )
        if method == "POST":
            return self._v3_auth_tokens(method, url, body, headers)
        raise NotImplementedError()

    def _v3_users(self, method, url, body, headers):
        if method == "GET":
            # list users
            body = self.fixtures.load("v3_users.json")
            return (
                httplib.OK,
                body,
                self.json_content_headers,
                httplib.responses[httplib.OK],
            )
        elif method == "POST":
            # create user
            body = self.fixtures.load("v3_create_user.json")
            return (
                httplib.CREATED,
                body,
                self.json_content_headers,
                httplib.responses[httplib.CREATED],
            )
        raise NotImplementedError()

    def _v3_users_a(self, method, url, body, headers):
        if method == "GET":
            # look up a user
            body = self.fixtures.load("v3_users_a.json")
            return (
                httplib.OK,
                body,
                self.json_content_headers,
                httplib.responses[httplib.OK],
            )
        if method == "PATCH":
            # enable / disable user
            body = self.fixtures.load("v3_users_a.json")
            return (
                httplib.OK,
                body,
                self.json_content_headers,
                httplib.responses[httplib.OK],
            )
        raise NotImplementedError()

    def _v3_users_b(self, method, url, body, headers):
        if method == "GET":
            # look up a user
            body = self.fixtures.load("v3_users_b.json")
            return (
                httplib.OK,
                body,
                self.json_content_headers,
                httplib.responses[httplib.OK],
            )
        raise NotImplementedError()

    def _v3_users_c(self, method, url, body, headers):
        if method == "GET":
            # look up a user
            body = self.fixtures.load("v3_users_c.json")
            return (
                httplib.OK,
                body,
                self.json_content_headers,
                httplib.responses[httplib.OK],
            )
        raise NotImplementedError()

    def _v3_roles(self, method, url, body, headers):
        if method == "GET":
            body = self.fixtures.load("v3_roles.json")
            return (
                httplib.OK,
                body,
                self.json_content_headers,
                httplib.responses[httplib.OK],
            )
        raise NotImplementedError()

    def _v3_domains_default_users_a_roles_a(self, method, url, body, headers):
        if method == "PUT":
            # grant domain role
            body = ""
            return (
                httplib.NO_CONTENT,
                body,
                self.json_content_headers,
                httplib.responses[httplib.NO_CONTENT],
            )
        elif method == "DELETE":
            # revoke domain role
            body = ""
            return (
                httplib.NO_CONTENT,
                body,
                self.json_content_headers,
                httplib.responses[httplib.NO_CONTENT],
            )
        raise NotImplementedError()

    def _v3_projects_a_users_a_roles_a(self, method, url, body, headers):
        if method == "PUT":
            # grant project role
            body = ""
            return (
                httplib.NO_CONTENT,
                body,
                self.json_content_headers,
                httplib.responses[httplib.NO_CONTENT],
            )
        elif method == "DELETE":
            # revoke project role
            body = ""
            return (
                httplib.NO_CONTENT,
                body,
                self.json_content_headers,
                httplib.responses[httplib.NO_CONTENT],
            )
        raise NotImplementedError()

    def _v3_domains_default(self, method, url, body, headers):
        if method == "GET":
            # get domain
            body = self.fixtures.load("v3_domains_default.json")
            return (
                httplib.OK,
                body,
                self.json_content_headers,
                httplib.responses[httplib.OK],
            )
        raise NotImplementedError()

    def _v3_users_a_projects(self, method, url, body, headers):
        if method == "GET":
            # get user projects
            body = self.fixtures.load("v3_users_a_projects.json")
            return (
                httplib.OK,
                body,
                self.json_content_headers,
                httplib.responses[httplib.OK],
            )
        raise NotImplementedError()

    def _v3_domains_default_users_a_roles(self, method, url, body, headers):
        if method == "GET":
            # get user domain roles
            body = self.fixtures.load("v3_domains_default_users_a_roles.json")
            return (
                httplib.OK,
                body,
                self.json_content_headers,
                httplib.responses[httplib.OK],
            )
        raise NotImplementedError()

    def _v3_OS_FEDERATION_identity_providers_idp_protocols_oidc_auth(
        self, method, url, body, headers
    ):
        if method == "GET":
            headers = self.json_content_headers.copy()
            headers["x-subject-token"] = "00000000000000000000000000000000"
            return (httplib.OK, body, headers, httplib.responses[httplib.OK])
        raise NotImplementedError()

    def _v3_OS_FEDERATION_projects(self, method, url, body, headers):
        if method == "GET":
            # get user projects
            body = json.dumps(
                {
                    "projects": [
                        {"id": "project_id", "name": "project_name"},
                        {"id": "project_id2", "name": "project_name2"},
                    ]
                }
            )
            return (
                httplib.OK,
                body,
                self.json_content_headers,
                httplib.responses[httplib.OK],
            )
        raise NotImplementedError()

    def _v3_auth_projects(self, method, url, body, headers):
        if method == "GET":
            # get user projects
            body = json.dumps(
                {
                    "projects": [
                        {"id": "project_id", "name": "project_name"},
                        {"id": "project_id2", "name": "project_name2"},
                    ]
                }
            )
            return (
                httplib.OK,
                body,
                self.json_content_headers,
                httplib.responses[httplib.OK],
            )
        raise NotImplementedError()


class OpenStackIdentity_3_0_AppCred_MockHttp(OpenStackIdentity_3_0_MockHttp):
    def _v3_auth_tokens(self, method, url, body, headers):
        if method == "POST":
            status = httplib.OK
            data = json.loads(body)
            if "application_credential" not in data["auth"]["identity"]["methods"]:
                status = httplib.UNAUTHORIZED
            else:
                appcred = data["auth"]["identity"]["application_credential"]
                if appcred["id"] != "appcred_id" or appcred["secret"] != "appcred_secret":
                    status = httplib.UNAUTHORIZED

            body = ComputeFileFixtures("openstack").load("_v3__auth.json")
            headers = self.json_content_headers.copy()
            headers["x-subject-token"] = "00000000000000000000000000000000"
            return (status, body, headers, httplib.responses[httplib.OK])
        raise NotImplementedError()


class OpenStackIdentity_3_0_federation_projects_MockHttp(OpenStackIdentity_3_0_MockHttp):
    fixtures = ComputeFileFixtures("openstack_identity/v3")
    json_content_headers = {"content-type": "application/json; charset=UTF-8"}

    def _v3_OS_FEDERATION_projects(self, method, url, body, headers):
        if method == "GET":
            # get user projects
            body = json.dumps({"projects": [{"id": "project_id"}]})
            return (
                httplib.OK,
                body,
                self.json_content_headers,
                httplib.responses[httplib.OK],
            )
        raise NotImplementedError()

    def _v3_auth_projects(self, method, url, body, headers):
        return (
            httplib.INTERNAL_SERVER_ERROR,
            body,
            self.json_content_headers,
            httplib.responses[httplib.INTERNAL_SERVER_ERROR],
        )


class OpenStackIdentity_2_0_Connection_VOMSMockHttp(MockHttp):
    fixtures = ComputeFileFixtures("openstack_identity/v2")
    json_content_headers = {"content-type": "application/json; charset=UTF-8"}

    def _v2_0_tokens(self, method, url, body, headers):
        if method == "POST":
            status = httplib.UNAUTHORIZED
            data = json.loads(body)
            if "voms" in data["auth"] and data["auth"]["voms"] is True:
                status = httplib.OK

            body = ComputeFileFixtures("openstack").load("_v2_0__auth.json")
            headers = self.json_content_headers.copy()
            headers["x-subject-token"] = "00000000000000000000000000000000"
            return (status, body, headers, httplib.responses[httplib.OK])
        raise NotImplementedError()

    def _v2_0_tenants(self, method, url, body, headers):
        if method == "GET":
            # get user projects
            body = json.dumps({"tenant": [{"name": "tenant_name"}]})
            return (
                httplib.OK,
                body,
                self.json_content_headers,
                httplib.responses[httplib.OK],
            )
        raise NotImplementedError()


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import base64
import os.path
import unittest

import libcloud.utils.files
from libcloud.test import MockHttp, make_response, generate_random_data
from libcloud.utils.py3 import b, httplib, urlparse
from libcloud.common.types import LibcloudError
from libcloud.storage.base import Object, Container
from libcloud.storage.types import (
    ObjectDoesNotExistError,
    ContainerIsNotEmptyError,
    ContainerDoesNotExistError,
    ContainerAlreadyExistsError,
)
from libcloud.test.file_fixtures import StorageFileFixtures
from libcloud.storage.drivers.atmos import AtmosDriver, AtmosConnection
from libcloud.storage.drivers.dummy import DummyIterator


class AtmosTests(unittest.TestCase):
    def setUp(self):
        AtmosDriver.connectionCls.conn_class = AtmosMockHttp
        AtmosDriver.path = ""
        AtmosMockHttp.type = None
        AtmosMockHttp.upload_created = False
        self.driver = AtmosDriver("dummy", base64.b64encode(b("dummy")))
        self._remove_test_file()

    def tearDown(self):
        self._remove_test_file()

    def _remove_test_file(self):
        file_path = os.path.abspath(__file__) + ".temp"

        try:
            os.unlink(file_path)
        except OSError:
            pass

    def test_list_containers(self):
        AtmosMockHttp.type = "EMPTY"
        containers = self.driver.list_containers()
        self.assertEqual(len(containers), 0)

        AtmosMockHttp.type = None
        containers = self.driver.list_containers()
        self.assertEqual(len(containers), 6)

    def test_list_container_objects(self):
        container = Container(name="test_container", extra={}, driver=self.driver)

        AtmosMockHttp.type = "EMPTY"
        objects = self.driver.list_container_objects(container=container)
        self.assertEqual(len(objects), 0)

        AtmosMockHttp.type = None
        objects = self.driver.list_container_objects(container=container)
        self.assertEqual(len(objects), 2)

        obj = [o for o in objects if o.name == "not-a-container1"][0]
        self.assertEqual(obj.meta_data["object_id"], "651eae32634bf84529c74eabd555fda48c7cead6")
        self.assertEqual(obj.container.name, "test_container")

    def test_get_container(self):
        container = self.driver.get_container(container_name="test_container")
        self.assertEqual(container.name, "test_container")
        self.assertEqual(container.extra["object_id"], "b21cb59a2ba339d1afdd4810010b0a5aba2ab6b9")

    def test_get_container_escaped(self):
        container = self.driver.get_container(container_name="test & container")
        self.assertEqual(container.name, "test & container")
        self.assertEqual(container.extra["object_id"], "b21cb59a2ba339d1afdd4810010b0a5aba2ab6b9")

    def test_get_container_not_found(self):
        try:
            self.driver.get_container(container_name="not_found")
        except ContainerDoesNotExistError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_create_container_success(self):
        container = self.driver.create_container(container_name="test_create_container")
        self.assertTrue(isinstance(container, Container))
        self.assertEqual(container.name, "test_create_container")
        self.assertEqual(container.extra["object_id"], "31a27b593629a3fe59f887fd973fd953e80062ce")

    def test_create_container_already_exists(self):
        AtmosMockHttp.type = "ALREADY_EXISTS"

        try:
            self.driver.create_container(container_name="test_create_container")
        except ContainerAlreadyExistsError:
            pass
        else:
            self.fail("Container already exists but an exception was not thrown")

    def test_delete_container_success(self):
        container = Container(name="foo_bar_container", extra={}, driver=self)
        result = self.driver.delete_container(container=container)
        self.assertTrue(result)

    def test_delete_container_not_found(self):
        AtmosMockHttp.type = "NOT_FOUND"
        container = Container(name="foo_bar_container", extra={}, driver=self)
        try:
            self.driver.delete_container(container=container)
        except ContainerDoesNotExistError:
            pass
        else:
            self.fail("Container does not exist but an exception was not thrown")

    def test_delete_container_not_empty(self):
        AtmosMockHttp.type = "NOT_EMPTY"
        container = Container(name="foo_bar_container", extra={}, driver=self)
        try:
            self.driver.delete_container(container=container)
        except ContainerIsNotEmptyError:
            pass
        else:
            self.fail("Container is not empty but an exception was not thrown")

    def test_get_object_success(self):
        obj = self.driver.get_object(container_name="test_container", object_name="test_object")
        self.assertEqual(obj.container.name, "test_container")
        self.assertEqual(obj.size, 555)
        self.assertEqual(obj.hash, "6b21c4a111ac178feacf9ec9d0c71f17")
        self.assertEqual(obj.extra["object_id"], "322dce3763aadc41acc55ef47867b8d74e45c31d6643")
        self.assertEqual(obj.extra["last_modified"], "Tue, 25 Jan 2011 22:01:49 GMT")
        self.assertEqual(obj.meta_data["foo-bar"], "test 1")
        self.assertEqual(obj.meta_data["bar-foo"], "test 2")

    def test_get_object_escaped(self):
        obj = self.driver.get_object(container_name="test & container", object_name="test & object")
        self.assertEqual(obj.container.name, "test & container")
        self.assertEqual(obj.size, 555)
        self.assertEqual(obj.hash, "6b21c4a111ac178feacf9ec9d0c71f17")
        self.assertEqual(obj.extra["object_id"], "322dce3763aadc41acc55ef47867b8d74e45c31d6643")
        self.assertEqual(obj.extra["last_modified"], "Tue, 25 Jan 2011 22:01:49 GMT")
        self.assertEqual(obj.meta_data["foo-bar"], "test 1")
        self.assertEqual(obj.meta_data["bar-foo"], "test 2")

    def test_get_object_not_found(self):
        try:
            self.driver.get_object(container_name="test_container", object_name="not_found")
        except ObjectDoesNotExistError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_delete_object_success(self):
        AtmosMockHttp.type = "DELETE"
        container = Container(name="foo_bar_container", extra={}, driver=self.driver)
        obj = Object(
            name="foo_bar_object",
            size=1000,
            hash=None,
            extra={},
            container=container,
            meta_data=None,
            driver=self.driver,
        )
        status = self.driver.delete_object(obj=obj)
        self.assertTrue(status)

    def test_delete_object_escaped_success(self):
        AtmosMockHttp.type = "DELETE"
        container = Container(name="foo & bar_container", extra={}, driver=self.driver)
        obj = Object(
            name="foo & bar_object",
            size=1000,
            hash=None,
            extra={},
            container=container,
            meta_data=None,
            driver=self.driver,
        )
        status = self.driver.delete_object(obj=obj)
        self.assertTrue(status)

    def test_delete_object_not_found(self):
        AtmosMockHttp.type = "NOT_FOUND"
        container = Container(name="foo_bar_container", extra={}, driver=self.driver)
        obj = Object(
            name="foo_bar_object",
            size=1000,
            hash=None,
            extra={},
            container=container,
            meta_data=None,
            driver=self.driver,
        )
        try:
            self.driver.delete_object(obj=obj)
        except ObjectDoesNotExistError:
            pass
        else:
            self.fail("Object does not exist but an exception was not thrown")

    def test_download_object_success(self):
        container = Container(name="foo_bar_container", extra={}, driver=self.driver)
        obj = Object(
            name="foo_bar_object",
            size=1000,
            hash=None,
            extra={},
            container=container,
            meta_data=None,
            driver=self.driver,
        )
        destination_path = os.path.abspath(__file__) + ".temp"
        result = self.driver.download_object(
            obj=obj,
            destination_path=destination_path,
            overwrite_existing=False,
            delete_on_failure=True,
        )
        self.assertTrue(result)

    def test_download_object_escaped_success(self):
        container = Container(name="foo & bar_container", extra={}, driver=self.driver)
        obj = Object(
            name="foo & bar_object",
            size=1000,
            hash=None,
            extra={},
            container=container,
            meta_data=None,
            driver=self.driver,
        )
        destination_path = os.path.abspath(__file__) + ".temp"
        result = self.driver.download_object(
            obj=obj,
            destination_path=destination_path,
            overwrite_existing=False,
            delete_on_failure=True,
        )
        self.assertTrue(result)

    def test_download_object_success_not_found(self):
        AtmosMockHttp.type = "NOT_FOUND"
        container = Container(name="foo_bar_container", extra={}, driver=self.driver)

        obj = Object(
            name="foo_bar_object",
            size=1000,
            hash=None,
            extra={},
            container=container,
            meta_data=None,
            driver=self.driver,
        )
        destination_path = os.path.abspath(__file__) + ".temp"
        try:
            self.driver.download_object(
                obj=obj,
                destination_path=destination_path,
                overwrite_existing=False,
                delete_on_failure=True,
            )
        except ObjectDoesNotExistError:
            pass
        else:
            self.fail("Object does not exist but an exception was not thrown")

    def test_download_object_as_stream(self):
        container = Container(name="foo_bar_container", extra={}, driver=self.driver)
        obj = Object(
            name="foo_bar_object",
            size=1000,
            hash=None,
            extra={},
            container=container,
            meta_data=None,
            driver=self.driver,
        )

        stream = self.driver.download_object_as_stream(obj=obj, chunk_size=None)
        self.assertTrue(hasattr(stream, "__iter__"))

    def test_download_object_as_stream_escaped(self):
        container = Container(name="foo & bar_container", extra={}, driver=self.driver)
        obj = Object(
            name="foo & bar_object",
            size=1000,
            hash=None,
            extra={},
            container=container,
            meta_data=None,
            driver=self.driver,
        )

        stream = self.driver.download_object_as_stream(obj=obj, chunk_size=None)
        self.assertTrue(hasattr(stream, "__iter__"))

    def test_upload_object_success(self):
        def upload_file(
            self,
            object_name=None,
            content_type=None,
            request_path=None,
            request_method=None,
            headers=None,
            file_path=None,
            stream=None,
        ):
            return {
                "response": make_response(
                    200, headers={"etag": "0cc175b9c0f1b6a831c399e269772661"}
                ),
                "bytes_transferred": 1000,
                "data_hash": "0cc175b9c0f1b6a831c399e269772661",
            }

        old_func = AtmosDriver._upload_object
        AtmosDriver._upload_object = upload_file

        path = os.path.abspath(__file__)
        container = Container(name="fbc", extra={}, driver=self)
        object_name = "ftu"
        extra = {"meta_data": {"some-value": "foobar"}}
        obj = self.driver.upload_object(
            file_path=path, container=container, extra=extra, object_name=object_name
        )
        self.assertEqual(obj.name, "ftu")
        self.assertEqual(obj.size, 1000)
        self.assertTrue("some-value" in obj.meta_data)
        AtmosDriver._upload_object = old_func

    def test_upload_object_no_content_type(self):
        def no_content_type(name):
            return None, None

        old_func = libcloud.utils.files.guess_file_mime_type
        libcloud.utils.files.guess_file_mime_type = no_content_type
        file_path = os.path.abspath(__file__)
        container = Container(name="fbc", extra={}, driver=self)
        object_name = "ftu"
        obj = self.driver.upload_object(
            file_path=file_path, container=container, object_name=object_name
        )

        # Just check that the file was uploaded OK, as the fallback
        # Content-Type header should be set (application/octet-stream).
        self.assertEqual(obj.name, object_name)
        libcloud.utils.files.guess_file_mime_type = old_func

    def test_upload_object_error(self):
        def dummy_content_type(name):
            return "application/zip", None

        def send(self, method, **kwargs):
            raise LibcloudError("")

        old_func1 = libcloud.utils.files.guess_file_mime_type
        libcloud.utils.files.guess_file_mime_type = dummy_content_type
        old_func2 = AtmosMockHttp.request
        AtmosMockHttp.request = send

        file_path = os.path.abspath(__file__)
        container = Container(name="fbc", extra={}, driver=self)
        object_name = "ftu"
        try:
            self.driver.upload_object(
                file_path=file_path, container=container, object_name=object_name
            )
        except LibcloudError:
            pass
        else:
            self.fail("Timeout while uploading but an exception was not thrown")
        finally:
            libcloud.utils.files.guess_file_mime_type = old_func1
            AtmosMockHttp.request = old_func2

    def test_upload_object_nonexistent_file(self):
        def dummy_content_type(name):
            return "application/zip", None

        old_func = libcloud.utils.files.guess_file_mime_type
        libcloud.utils.files.guess_file_mime_type = dummy_content_type

        file_path = os.path.abspath(__file__ + ".inexistent")
        container = Container(name="fbc", extra={}, driver=self)
        object_name = "ftu"
        try:
            self.driver.upload_object(
                file_path=file_path, container=container, object_name=object_name
            )
        except OSError:
            pass
        else:
            self.fail("Inesitent but an exception was not thrown")
        finally:
            libcloud.utils.files.guess_file_mime_type = old_func

    def test_upload_object_via_stream_new_object(self):
        def dummy_content_type(name):
            return "application/zip", None

        old_func = libcloud.utils.files.guess_file_mime_type
        libcloud.utils.files.guess_file_mime_type = dummy_content_type

        container = Container(name="fbc", extra={}, driver=self)
        object_name = "ftsdn"
        iterator = DummyIterator(data=["2", "3", "5"])
        try:
            self.driver.upload_object_via_stream(
                container=container, object_name=object_name, iterator=iterator
            )
        finally:
            libcloud.utils.files.guess_file_mime_type = old_func

    def test_upload_object_via_stream_existing_object(self):
        def dummy_content_type(name):
            return "application/zip", None

        old_func = libcloud.utils.files.guess_file_mime_type
        libcloud.utils.files.guess_file_mime_type = dummy_content_type

        container = Container(name="fbc", extra={}, driver=self)
        object_name = "ftsde"
        iterator = DummyIterator(data=["2", "3", "5"])
        try:
            self.driver.upload_object_via_stream(
                container=container, object_name=object_name, iterator=iterator
            )
        finally:
            libcloud.utils.files.guess_file_mime_type = old_func

    def test_upload_object_via_stream_no_content_type(self):
        def no_content_type(name):
            return None, None

        old_func = libcloud.utils.files.guess_file_mime_type
        libcloud.utils.files.guess_file_mime_type = no_content_type

        container = Container(name="fbc", extra={}, driver=self)
        object_name = "ftsdct"
        iterator = DummyIterator(data=["2", "3", "5"])
        try:
            self.driver.upload_object_via_stream(
                container=container, object_name=object_name, iterator=iterator
            )
        except AttributeError:
            pass
        else:
            self.fail("File content type not provided" " but an exception was not thrown")
        finally:
            libcloud.utils.files.guess_file_mime_type = old_func

    def test_signature_algorithm(self):
        test_uid = "fredsmagicuid"
        test_key = base64.b64encode(b("ssssshhhhhmysecretkey"))
        test_date = "Mon, 04 Jul 2011 07:39:19 GMT"
        test_values = [
            ("GET", "/rest/namespace/foo", "", {}, "WfSASIA25TuqO2n0aO9k/dtg6S0="),
            (
                "GET",
                "/rest/namespace/foo%20%26%20bar",
                "",
                {},
                "vmlqXqcInxxoP4YX5mR09BonjX4=",
            ),
            ("POST", "/rest/namespace/foo", "", {}, "oYKdsF+1DOuUT7iX5CJCDym2EQk="),
            ("PUT", "/rest/namespace/foo", "", {}, "JleF9dpSWhaT3B2swZI3s41qqs4="),
            ("DELETE", "/rest/namespace/foo", "", {}, "2IX+Bd5XZF5YY+g4P59qXV1uLpo="),
            (
                "GET",
                "/rest/namespace/foo?metata/system",
                "",
                {},
                "zuHDEAgKM1winGnWn3WBsqnz4ks=",
            ),
            (
                "POST",
                "/rest/namespace/foo?metadata/user",
                "",
                {"x-emc-meta": "fakemeta=fake, othermeta=faketoo"},
                "7sLx1nxPIRAtocfv02jz9h1BjbU=",
            ),
        ]

        class FakeDriver:
            path = ""

        for method, action, api_path, headers, expected in test_values:
            c = AtmosConnection(test_uid, test_key)
            c.method = method
            c.action = action
            d = FakeDriver()
            d.path = api_path
            c.driver = d
            headers = c.add_default_headers(headers)
            headers["Date"] = headers["x-emc-date"] = test_date
            self.assertEqual(c._calculate_signature({}, headers), b(expected).decode("utf-8"))


class AtmosMockHttp(MockHttp, unittest.TestCase):
    fixtures = StorageFileFixtures("atmos")
    upload_created = False
    upload_stream_created = False

    def __init__(self, *args, **kwargs):
        unittest.TestCase.__init__(self)

        if kwargs.get("host", None) and kwargs.get("port", None):
            MockHttp.__init__(self, *args, **kwargs)

        self._upload_object_via_stream_first_request = True

    def runTest(self):
        pass

    def request(self, method, url, body=None, headers=None, raw=False, stream=False):
        headers = headers or {}
        parsed = urlparse.urlparse(url)
        if parsed.query.startswith("metadata/"):
            parsed = list(parsed)
            parsed[2] = parsed[2] + "/" + parsed[4]
            parsed[4] = ""
            url = urlparse.urlunparse(parsed)
        return super().request(method, url, body, headers, raw)

    def _rest_namespace_EMPTY(self, method, url, body, headers):
        body = self.fixtures.load("empty_directory_listing.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _rest_namespace(self, method, url, body, headers):
        body = self.fixtures.load("list_containers.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _rest_namespace_test_container_EMPTY(self, method, url, body, headers):
        body = self.fixtures.load("empty_directory_listing.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _rest_namespace_test_container(self, method, url, body, headers):
        body = self.fixtures.load("list_containers.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _rest_namespace_test_container__metadata_system(self, method, url, body, headers):
        headers = {"x-emc-meta": "objectid=b21cb59a2ba339d1afdd4810010b0a5aba2ab6b9"}
        return (httplib.OK, "", headers, httplib.responses[httplib.OK])

    def _rest_namespace_test_20_26_20container__metadata_system(self, method, url, body, headers):
        headers = {"x-emc-meta": "objectid=b21cb59a2ba339d1afdd4810010b0a5aba2ab6b9"}
        return (httplib.OK, "", headers, httplib.responses[httplib.OK])

    def _rest_namespace_not_found__metadata_system(self, method, url, body, headers):
        body = self.fixtures.load("not_found.xml")
        return (httplib.NOT_FOUND, body, {}, httplib.responses[httplib.NOT_FOUND])

    def _rest_namespace_test_create_container(self, method, url, body, headers):
        return (httplib.OK, "", {}, httplib.responses[httplib.OK])

    def _rest_namespace_test_create_container__metadata_system(self, method, url, body, headers):
        headers = {"x-emc-meta": "objectid=31a27b593629a3fe59f887fd973fd953e80062ce"}
        return (httplib.OK, "", headers, httplib.responses[httplib.OK])

    def _rest_namespace_test_create_container_ALREADY_EXISTS(self, method, url, body, headers):
        body = self.fixtures.load("already_exists.xml")
        return (httplib.BAD_REQUEST, body, {}, httplib.responses[httplib.BAD_REQUEST])

    def _rest_namespace_foo_bar_container(self, method, url, body, headers):
        return (httplib.OK, "", {}, httplib.responses[httplib.OK])

    def _rest_namespace_foo_bar_container_NOT_FOUND(self, method, url, body, headers):
        body = self.fixtures.load("not_found.xml")
        return (httplib.NOT_FOUND, body, {}, httplib.responses[httplib.NOT_FOUND])

    def _rest_namespace_foo_bar_container_NOT_EMPTY(self, method, url, body, headers):
        body = self.fixtures.load("not_empty.xml")
        return (httplib.BAD_REQUEST, body, {}, httplib.responses[httplib.BAD_REQUEST])

    def _rest_namespace_test_container_test_object_metadata_system(
        self, method, url, body, headers
    ):
        meta = {
            "objectid": "322dce3763aadc41acc55ef47867b8d74e45c31d6643",
            "size": "555",
            "mtime": "2011-01-25T22:01:49Z",
        }
        headers = {"x-emc-meta": ", ".join([k + "=" + v for k, v in list(meta.items())])}
        return (httplib.OK, "", headers, httplib.responses[httplib.OK])

    def _rest_namespace_test_20_26_20container_test_20_26_20object_metadata_system(
        self, method, url, body, headers
    ):
        meta = {
            "objectid": "322dce3763aadc41acc55ef47867b8d74e45c31d6643",
            "size": "555",
            "mtime": "2011-01-25T22:01:49Z",
        }
        headers = {"x-emc-meta": ", ".join([k + "=" + v for k, v in list(meta.items())])}
        return (httplib.OK, "", headers, httplib.responses[httplib.OK])

    def _rest_namespace_test_container_test_object_metadata_user(self, method, url, body, headers):
        meta = {
            "md5": "6b21c4a111ac178feacf9ec9d0c71f17",
            "foo-bar": "test 1",
            "bar-foo": "test 2",
        }
        headers = {"x-emc-meta": ", ".join([k + "=" + v for k, v in list(meta.items())])}
        return (httplib.OK, "", headers, httplib.responses[httplib.OK])

    def _rest_namespace_test_20_26_20container_test_20_26_20object_metadata_user(
        self, method, url, body, headers
    ):
        meta = {
            "md5": "6b21c4a111ac178feacf9ec9d0c71f17",
            "foo-bar": "test 1",
            "bar-foo": "test 2",
        }
        headers = {"x-emc-meta": ", ".join([k + "=" + v for k, v in list(meta.items())])}
        return (httplib.OK, "", headers, httplib.responses[httplib.OK])

    def _rest_namespace_test_container_not_found_metadata_system(self, method, url, body, headers):
        body = self.fixtures.load("not_found.xml")
        return (httplib.NOT_FOUND, body, {}, httplib.responses[httplib.NOT_FOUND])

    def _rest_namespace_foo_bar_container_foo_bar_object_DELETE(self, method, url, body, headers):
        return (httplib.OK, "", {}, httplib.responses[httplib.OK])

    def _rest_namespace_foo_20_26_20bar_container_foo_20_26_20bar_object_DELETE(
        self, method, url, body, headers
    ):
        return (httplib.OK, "", {}, httplib.responses[httplib.OK])

    def _rest_namespace_foo_bar_container_foo_bar_object_NOT_FOUND(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("not_found.xml")
        return (httplib.NOT_FOUND, body, {}, httplib.responses[httplib.NOT_FOUND])

    def _rest_namespace_fbc_ftu_metadata_system(self, method, url, body, headers):
        if not self.upload_created:
            self.__class__.upload_created = True
            body = self.fixtures.load("not_found.xml")
            return (httplib.NOT_FOUND, body, {}, httplib.responses[httplib.NOT_FOUND])

        self.__class__.upload_created = False
        meta = {
            "objectid": "322dce3763aadc41acc55ef47867b8d74e45c31d6643",
            "size": "555",
            "mtime": "2011-01-25T22:01:49Z",
        }
        headers = {"x-emc-meta": ", ".join([k + "=" + v for k, v in list(meta.items())])}
        return (httplib.OK, "", headers, httplib.responses[httplib.OK])

    def _rest_namespace_fbc_ftu_metadata_user(self, method, url, body, headers):
        self.assertTrue("x-emc-meta" in headers)
        return (httplib.OK, "", {}, httplib.responses[httplib.OK])

    def _rest_namespace_fbc_ftsdn_metadata_system(self, method, url, body, headers):
        if not self.upload_stream_created:
            self.__class__.upload_stream_created = True
            body = self.fixtures.load("not_found.xml")
            return (httplib.NOT_FOUND, body, {}, httplib.responses[httplib.NOT_FOUND])

        self.__class__.upload_stream_created = False
        meta = {
            "objectid": "322dce3763aadc41acc55ef47867b8d74e45c31d6643",
            "size": "555",
            "mtime": "2011-01-25T22:01:49Z",
        }
        headers = {"x-emc-meta": ", ".join([k + "=" + v for k, v in list(meta.items())])}
        return (httplib.OK, "", headers, httplib.responses[httplib.OK])

    def _rest_namespace_fbc_ftsdn(self, method, url, body, headers):
        if self._upload_object_via_stream_first_request:
            self.assertTrue("Range" not in headers)
            self.assertEqual(method, "POST")
            self._upload_object_via_stream_first_request = False
        else:
            self.assertTrue("Range" in headers)
            self.assertEqual(method, "PUT")
        return (httplib.OK, "", {}, httplib.responses[httplib.OK])

    def _rest_namespace_fbc_ftsdn_metadata_user(self, method, url, body, headers):
        self.assertTrue("x-emc-meta" in headers)
        return (httplib.OK, "", {}, httplib.responses[httplib.OK])

    def _rest_namespace_fbc_ftsde_metadata_system(self, method, url, body, headers):
        meta = {
            "objectid": "322dce3763aadc41acc55ef47867b8d74e45c31d6643",
            "size": "555",
            "mtime": "2011-01-25T22:01:49Z",
        }
        headers = {"x-emc-meta": ", ".join([k + "=" + v for k, v in list(meta.items())])}
        return (httplib.OK, "", headers, httplib.responses[httplib.OK])

    def _rest_namespace_fbc_ftsde(self, method, url, body, headers):
        if self._upload_object_via_stream_first_request:
            self.assertTrue("Range" not in headers)
            self._upload_object_via_stream_first_request = False
        else:
            self.assertTrue("Range" in headers)
        self.assertEqual(method, "PUT")
        return (httplib.OK, "", {}, httplib.responses[httplib.OK])

    def _rest_namespace_fbc_ftsde_metadata_user(self, method, url, body, headers):
        self.assertTrue("x-emc-meta" in headers)
        return (httplib.OK, "", {}, httplib.responses[httplib.OK])

    def _rest_namespace_fbc_ftsd_metadata_system(self, method, url, body, headers):
        meta = {
            "objectid": "322dce3763aadc41acc55ef47867b8d74e45c31d6643",
            "size": "555",
            "mtime": "2011-01-25T22:01:49Z",
        }
        headers = {"x-emc-meta": ", ".join([k + "=" + v for k, v in list(meta.items())])}
        return (httplib.OK, "", headers, httplib.responses[httplib.OK])

    def _rest_namespace_foo_bar_container_foo_bar_object(self, method, url, body, headers):
        body = generate_random_data(1000)
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _rest_namespace_foo_20_26_20bar_container_foo_20_26_20bar_object(
        self, method, url, body, headers
    ):
        body = generate_random_data(1000)
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _rest_namespace_fbc_ftu(self, method, url, body, headers):
        return (httplib.CREATED, "", {}, httplib.responses[httplib.CREATED])


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
import base64
import unittest

from libcloud.test import MockHttp
from libcloud.utils.py3 import b, httplib
from libcloud.common.types import InvalidCredsError
from libcloud.test.compute import TestCaseMixin
from libcloud.test.secrets import BRIGHTBOX_PARAMS
from libcloud.compute.types import NodeState
from libcloud.test.file_fixtures import ComputeFileFixtures
from libcloud.compute.drivers.brightbox import BrightboxNodeDriver

try:
    import simplejson as json
except ImportError:
    import json


USER_DATA = "#!/bin/sh\ntest_script.sh\n"


class BrightboxTest(unittest.TestCase, TestCaseMixin):
    def setUp(self):
        BrightboxNodeDriver.connectionCls.conn_class = BrightboxMockHttp
        BrightboxMockHttp.type = None
        BrightboxNodeDriver.connectionCls.token = "test"
        self.driver = BrightboxNodeDriver(*BRIGHTBOX_PARAMS)

    def test_authentication(self):
        BrightboxMockHttp.type = "INVALID_CLIENT"
        self.assertRaises(InvalidCredsError, self.driver.list_nodes)

        BrightboxMockHttp.type = "UNAUTHORIZED_CLIENT"
        self.assertRaises(InvalidCredsError, self.driver.list_nodes)

    def test_invalid_api_version(self):
        kwargs = {"api_version": "2.0"}
        self.driver = BrightboxNodeDriver(*BRIGHTBOX_PARAMS, **kwargs)
        self.assertRaises(Exception, self.driver.list_locations)

    def test_other_host(self):
        kwargs = {"host": "api.gbt.brightbox.com"}
        self.driver = BrightboxNodeDriver(*BRIGHTBOX_PARAMS, **kwargs)
        locations = self.driver.list_locations()
        self.assertEqual(len(locations), 0)

    def test_list_nodes(self):
        nodes = self.driver.list_nodes()
        self.assertEqual(len(nodes), 2)
        self.assertEqual(len(nodes[0].public_ips), 1)
        self.assertEqual(len(nodes[1].public_ips), 1)
        self.assertEqual(len(nodes[0].private_ips), 1)
        self.assertEqual(len(nodes[1].private_ips), 1)
        self.assertTrue("109.107.35.16" in nodes[0].public_ips)
        self.assertTrue("10.74.210.210" in nodes[0].private_ips)
        self.assertTrue("10.240.228.234" in nodes[1].private_ips)
        self.assertTrue("2a02:1348:14c:393a:24:19ff:fef0:e4ea" in nodes[1].public_ips)
        self.assertEqual(nodes[0].state, NodeState.RUNNING)
        self.assertEqual(nodes[1].state, NodeState.RUNNING)

    def test_list_node_extras(self):
        nodes = self.driver.list_nodes()
        self.assertFalse(nodes[0].size is None)
        self.assertFalse(nodes[1].size is None)
        self.assertFalse(nodes[0].image is None)
        self.assertFalse(nodes[1].image is None)
        self.assertEqual(nodes[0].image.id, "img-arm8f")
        self.assertEqual(nodes[0].size.id, "typ-urtky")
        self.assertEqual(nodes[1].image.id, "img-j93gd")
        self.assertEqual(nodes[1].size.id, "typ-qdiwq")
        self.assertEqual(nodes[0].extra["fqdn"], "srv-xvpn7.gb1.brightbox.com")
        self.assertEqual(nodes[1].extra["fqdn"], "srv-742vn.gb1.brightbox.com")
        self.assertEqual(nodes[0].extra["hostname"], "srv-xvpn7")
        self.assertEqual(nodes[1].extra["hostname"], "srv-742vn")
        self.assertEqual(nodes[0].extra["status"], "active")
        self.assertEqual(nodes[1].extra["status"], "active")
        self.assertTrue("interfaces" in nodes[0].extra)
        self.assertTrue("zone" in nodes[0].extra)
        self.assertTrue("snapshots" in nodes[0].extra)
        self.assertTrue("server_groups" in nodes[0].extra)
        self.assertTrue("started_at" in nodes[0].extra)
        self.assertTrue("created_at" in nodes[0].extra)
        self.assertFalse("deleted_at" in nodes[0].extra)

    def test_list_sizes(self):
        sizes = self.driver.list_sizes()
        self.assertEqual(len(sizes), 7)
        self.assertEqual(sizes[0].id, "typ-4nssg")
        self.assertEqual(sizes[0].name, "Brightbox Nano Instance")
        self.assertEqual(sizes[0].ram, 512)
        self.assertEqual(sizes[0].disk, 20480)
        self.assertEqual(sizes[0].bandwidth, 0)
        self.assertEqual(sizes[0].price, 0)

    def test_list_images(self):
        images = self.driver.list_images()
        self.assertEqual(len(images), 3)
        self.assertEqual(images[0].id, "img-99q79")
        self.assertEqual(images[0].name, "CentOS 5.5 server")
        self.assertTrue("ancestor" in images[0].extra)
        self.assertFalse("licence_name" in images[0].extra)

    def test_list_images_extras(self):
        images = self.driver.list_images()
        extra = images[-1].extra
        self.assertEqual(extra["arch"], "i686")
        self.assertFalse(extra["compatibility_mode"])
        self.assertEqual(extra["created_at"], "2012-01-22T05:36:24Z")
        self.assertTrue("description" in extra)
        self.assertEqual(extra["disk_size"], 671)
        self.assertFalse("min_ram" in extra)
        self.assertFalse(extra["official"])
        self.assertEqual(extra["owner"], "acc-tqs4c")
        self.assertTrue(extra["public"])
        self.assertEqual(extra["source"], "oneiric-i386-20178.gz")
        self.assertEqual(extra["source_type"], "upload")
        self.assertEqual(extra["status"], "deprecated")
        self.assertEqual(extra["username"], "ubuntu")
        self.assertEqual(extra["virtual_size"], 1025)
        self.assertFalse("ancestor" in extra)
        self.assertFalse("licence_name" in extra)

    def test_list_locations(self):
        locations = self.driver.list_locations()
        self.assertEqual(locations[0].id, "zon-6mxqw")
        self.assertEqual(locations[0].name, "gb1-a")
        self.assertEqual(locations[1].id, "zon-remk1")
        self.assertEqual(locations[1].name, "gb1-b")

    def test_reboot_node_response(self):
        node = self.driver.list_nodes()[0]
        self.assertRaises(NotImplementedError, self.driver.reboot_node, [node])

    def test_destroy_node(self):
        node = self.driver.list_nodes()[0]
        self.assertTrue(self.driver.destroy_node(node))

    def test_create_node(self):
        size = self.driver.list_sizes()[0]
        image = self.driver.list_images()[0]
        node = self.driver.create_node(name="Test Node", image=image, size=size)
        self.assertEqual("srv-p61uj", node.id)
        self.assertEqual("Test Node", node.name)
        self.assertEqual("gb1-a", node.extra["zone"].name)

    def test_create_node_in_location(self):
        size = self.driver.list_sizes()[0]
        image = self.driver.list_images()[0]
        location = self.driver.list_locations()[1]
        node = self.driver.create_node(name="Test Node", image=image, size=size, location=location)
        self.assertEqual("srv-nnumd", node.id)
        self.assertEqual("Test Node", node.name)
        self.assertEqual("gb1-b", node.extra["zone"].name)

    def test_create_node_with_user_data(self):
        size = self.driver.list_sizes()[0]
        image = self.driver.list_images()[0]
        node = self.driver.create_node(
            name="Test Node", image=image, size=size, ex_userdata=USER_DATA
        )
        decoded = base64.b64decode(b(node.extra["user_data"])).decode("ascii")
        self.assertEqual("gb1-a", node.extra["zone"].name)
        self.assertEqual(USER_DATA, decoded)

    def test_create_node_with_a_server_group(self):
        size = self.driver.list_sizes()[0]
        image = self.driver.list_images()[0]
        node = self.driver.create_node(
            name="Test Node", image=image, size=size, ex_servergroup="grp-12345"
        )
        self.assertEqual("gb1-a", node.extra["zone"].name)
        self.assertEqual(len(node.extra["server_groups"]), 1)
        self.assertEqual(node.extra["server_groups"][0]["id"], "grp-12345")

    def test_create_node_with_a_list_of_server_groups(self):
        size = self.driver.list_sizes()[0]
        image = self.driver.list_images()[0]
        node = self.driver.create_node(
            name="Test Node",
            image=image,
            size=size,
            ex_servergroup=["grp-12345", "grp-67890"],
        )
        self.assertEqual("gb1-a", node.extra["zone"].name)
        self.assertEqual(len(node.extra["server_groups"]), 2)
        self.assertEqual(node.extra["server_groups"][0]["id"], "grp-12345")
        self.assertEqual(node.extra["server_groups"][1]["id"], "grp-67890")

    def test_list_cloud_ips(self):
        cip_list = self.driver.ex_list_cloud_ips()
        self.assertEqual(len(cip_list), 4)
        self.assertEqual(cip_list[2]["status"], "mapped")
        cip_check = cip_list[0]
        self.assertEqual(cip_check["id"], "cip-tlrp3")
        self.assertEqual(cip_check["public_ip"], "109.107.35.16")
        self.assertEqual(cip_check["reverse_dns"], "cip-109-107-35-16.gb1.brightbox.com")
        self.assertEqual(cip_check["status"], "unmapped")
        self.assertTrue(cip_check["server"] is None)
        self.assertTrue(cip_check["server_group"] is None)
        self.assertTrue(cip_check["interface"] is None)
        self.assertTrue(cip_check["load_balancer"] is None)

    def test_create_cloud_ip(self):
        cip = self.driver.ex_create_cloud_ip()
        self.assertEqual(cip["id"], "cip-jsjc5")
        self.assertEqual(cip["reverse_dns"], "cip-109-107-37-234.gb1.brightbox.com")

    def test_create_cloud_ip_with_dns(self):
        cip = self.driver.ex_create_cloud_ip("fred.co.uk")
        self.assertEqual(cip["id"], "cip-jsjc5")
        self.assertEqual(cip["reverse_dns"], "fred.co.uk")

    def test_map_cloud_ip(self):
        self.assertTrue(self.driver.ex_map_cloud_ip("cip-jsjc5", "int-ztqbx"))

    def test_unmap_cloud_ip(self):
        self.assertTrue(self.driver.ex_unmap_cloud_ip("cip-jsjc5"))

    def test_update_cloud_ip(self):
        self.assertTrue(self.driver.ex_update_cloud_ip("cip-jsjc5", "fred.co.uk"))

    def test_destroy_cloud_ip(self):
        self.assertTrue(self.driver.ex_destroy_cloud_ip("cip-jsjc5"))


class BrightboxMockHttp(MockHttp):
    fixtures = ComputeFileFixtures("brightbox")

    def _token(self, method, url, body, headers):
        if method == "POST":
            return self.test_response(httplib.OK, self.fixtures.load("token.json"))

    def _token_INVALID_CLIENT(self, method, url, body, headers):
        if method == "POST":
            return self.test_response(httplib.BAD_REQUEST, '{"error":"invalid_client"}')

    def _token_UNAUTHORIZED_CLIENT(self, method, url, body, headers):
        if method == "POST":
            return self.test_response(httplib.UNAUTHORIZED, '{"error":"unauthorized_client"}')

    def _1_0_servers_INVALID_CLIENT(self, method, url, body, headers):
        return self.test_response(httplib.BAD_REQUEST, '{"error":"invalid_client"}')

    def _1_0_servers_UNAUTHORIZED_CLIENT(self, method, url, body, headers):
        return self.test_response(httplib.UNAUTHORIZED, '{"error":"unauthorized_client"}')

    def _1_0_images(self, method, url, body, headers):
        if method == "GET":
            return self.test_response(httplib.OK, self.fixtures.load("list_images.json"))

    def _1_0_servers(self, method, url, body, headers):
        if method == "GET":
            return self.test_response(httplib.OK, self.fixtures.load("list_servers.json"))
        elif method == "POST":
            body = json.loads(body)
            encoded = base64.b64encode(b(USER_DATA)).decode("ascii")

            if "user_data" in body and body["user_data"] != encoded:
                data = (
                    '{"error_name":"dodgy user data", "errors": ["User data not encoded properly"]}'
                )
                return self.test_response(httplib.BAD_REQUEST, data)
            if body.get("zone", "") == "zon-remk1":
                node = json.loads(self.fixtures.load("create_server_gb1_b.json"))
            else:
                node = json.loads(self.fixtures.load("create_server_gb1_a.json"))
            node["name"] = body["name"]
            if "server_groups" in body:
                node["server_groups"] = [{"id": x} for x in body["server_groups"]]
            if "user_data" in body:
                node["user_data"] = body["user_data"]
            return self.test_response(httplib.ACCEPTED, json.dumps(node))

    def _1_0_servers_srv_xvpn7(self, method, url, body, headers):
        if method == "DELETE":
            return self.test_response(httplib.ACCEPTED, "")

    def _1_0_server_types(self, method, url, body, headers):
        if method == "GET":
            return self.test_response(httplib.OK, self.fixtures.load("list_server_types.json"))

    def _1_0_zones(self, method, url, body, headers):
        if method == "GET":
            if headers["Host"] == "api.gbt.brightbox.com":
                return self.test_response(httplib.OK, "{}")
            else:
                return self.test_response(httplib.OK, self.fixtures.load("list_zones.json"))

    def _2_0_zones(self, method, url, body, headers):
        data = '{"error_name":"unrecognised_endpoint", "errors": ["The request was for an unrecognised API endpoint"]}'
        return self.test_response(httplib.BAD_REQUEST, data)

    def _1_0_cloud_ips(self, method, url, body, headers):
        if method == "GET":
            return self.test_response(httplib.OK, self.fixtures.load("list_cloud_ips.json"))
        elif method == "POST":
            if body:
                body = json.loads(body)

            node = json.loads(self.fixtures.load("create_cloud_ip.json"))

            if "reverse_dns" in body:
                node["reverse_dns"] = body["reverse_dns"]
            return self.test_response(httplib.ACCEPTED, json.dumps(node))

    def _1_0_cloud_ips_cip_jsjc5(self, method, url, body, headers):
        if method == "DELETE":
            return self.test_response(httplib.OK, "")
        elif method == "PUT":
            body = json.loads(body)
            if body.get("reverse_dns", None) == "fred.co.uk":
                return self.test_response(httplib.OK, "")
            else:
                return self.test_response(
                    httplib.BAD_REQUEST,
                    '{"error_name":"bad dns", "errors": ["Bad dns"]}',
                )

    def _1_0_cloud_ips_cip_jsjc5_map(self, method, url, body, headers):
        if method == "POST":
            body = json.loads(body)
            if "destination" in body:
                return self.test_response(httplib.ACCEPTED, "")
            else:
                data = '{"error_name":"bad destination", "errors": ["Bad destination"]}'
                return self.test_response(httplib.BAD_REQUEST, data)

    def _1_0_cloud_ips_cip_jsjc5_unmap(self, method, url, body, headers):
        if method == "POST":
            return self.test_response(httplib.ACCEPTED, "")

    def test_response(self, status, body):
        return (
            status,
            body,
            {"content-type": "application/json"},
            httplib.responses[status],
        )


if __name__ == "__main__":
    sys.exit(unittest.main())

# vim: autoindent tabstop=4 shiftwidth=4 expandtab softtabstop=4
# filetype=python



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Provider related utilities
"""


from types import ModuleType
from typing import TYPE_CHECKING, Type, Union

from libcloud.compute.types import OLD_CONSTANT_TO_NEW_MAPPING, Provider
from libcloud.common.providers import get_driver as _get_provider_driver
from libcloud.common.providers import set_driver as _set_provider_driver
from libcloud.compute.deprecated import DEPRECATED_DRIVERS

if TYPE_CHECKING:
    # NOTE: This is needed to avoid having setup.py depend on requests
    from libcloud.compute.base import NodeDriver

__all__ = ["Provider", "DRIVERS", "get_driver"]

DRIVERS = {
    Provider.AZURE: ("libcloud.compute.drivers.azure", "AzureNodeDriver"),
    Provider.AZURE_ARM: ("libcloud.compute.drivers.azure_arm", "AzureNodeDriver"),
    Provider.DUMMY: ("libcloud.compute.drivers.dummy", "DummyNodeDriver"),
    Provider.EC2: ("libcloud.compute.drivers.ec2", "EC2NodeDriver"),
    Provider.CLOUDSIGMA: (
        "libcloud.compute.drivers.cloudsigma",
        "CloudSigmaNodeDriver",
    ),
    Provider.GCE: ("libcloud.compute.drivers.gce", "GCENodeDriver"),
    Provider.RACKSPACE: ("libcloud.compute.drivers.rackspace", "RackspaceNodeDriver"),
    Provider.RACKSPACE_FIRST_GEN: (
        "libcloud.compute.drivers.rackspace",
        "RackspaceFirstGenNodeDriver",
    ),
    Provider.VPSNET: ("libcloud.compute.drivers.vpsnet", "VPSNetNodeDriver"),
    Provider.LINODE: ("libcloud.compute.drivers.linode", "LinodeNodeDriver"),
    Provider.RIMUHOSTING: (
        "libcloud.compute.drivers.rimuhosting",
        "RimuHostingNodeDriver",
    ),
    Provider.EUCALYPTUS: ("libcloud.compute.drivers.ec2", "EucNodeDriver"),
    Provider.OPENNEBULA: (
        "libcloud.compute.drivers.opennebula",
        "OpenNebulaNodeDriver",
    ),
    Provider.BRIGHTBOX: ("libcloud.compute.drivers.brightbox", "BrightboxNodeDriver"),
    Provider.NIMBUS: ("libcloud.compute.drivers.ec2", "NimbusNodeDriver"),
    Provider.GANDI: ("libcloud.compute.drivers.gandi", "GandiNodeDriver"),
    Provider.DIMENSIONDATA: (
        "libcloud.compute.drivers.dimensiondata",
        "DimensionDataNodeDriver",
    ),
    Provider.OPENSTACK: ("libcloud.compute.drivers.openstack", "OpenStackNodeDriver"),
    Provider.VCLOUD: ("libcloud.compute.drivers.vcloud", "VCloudNodeDriver"),
    Provider.TERREMARK: ("libcloud.compute.drivers.vcloud", "TerremarkDriver"),
    Provider.CLOUDSTACK: (
        "libcloud.compute.drivers.cloudstack",
        "CloudStackNodeDriver",
    ),
    Provider.LIBVIRT: ("libcloud.compute.drivers.libvirt_driver", "LibvirtNodeDriver"),
    Provider.VCL: ("libcloud.compute.drivers.vcl", "VCLNodeDriver"),
    Provider.KTUCLOUD: ("libcloud.compute.drivers.ktucloud", "KTUCloudNodeDriver"),
    Provider.ABIQUO: ("libcloud.compute.drivers.abiquo", "AbiquoNodeDriver"),
    Provider.DIGITAL_OCEAN: (
        "libcloud.compute.drivers.digitalocean",
        "DigitalOceanNodeDriver",
    ),
    Provider.EXOSCALE: ("libcloud.compute.drivers.exoscale", "ExoscaleNodeDriver"),
    Provider.IKOULA: ("libcloud.compute.drivers.ikoula", "IkoulaNodeDriver"),
    Provider.OUTSCALE_SAS: ("libcloud.compute.drivers.ec2", "OutscaleSASNodeDriver"),
    Provider.OUTSCALE_INC: ("libcloud.compute.drivers.ec2", "OutscaleINCNodeDriver"),
    Provider.OUTSCALE: ("libcloud.compute.drivers.outscale", "OutscaleNodeDriver"),
    Provider.VSPHERE: ("libcloud.compute.drivers.vsphere", "VSphereNodeDriver"),
    Provider.VULTR: ("libcloud.compute.drivers.vultr", "VultrNodeDriver"),
    Provider.AURORACOMPUTE: (
        "libcloud.compute.drivers.auroracompute",
        "AuroraComputeNodeDriver",
    ),
    Provider.EQUINIXMETAL: (
        "libcloud.compute.drivers.equinixmetal",
        "EquinixMetalNodeDriver",
    ),
    Provider.ONAPP: ("libcloud.compute.drivers.onapp", "OnAppNodeDriver"),
    Provider.OVH: ("libcloud.compute.drivers.ovh", "OvhNodeDriver"),
    Provider.INTERNETSOLUTIONS: (
        "libcloud.compute.drivers.internetsolutions",
        "InternetSolutionsNodeDriver",
    ),
    Provider.NTTA: ("libcloud.compute.drivers.ntta", "NTTAmericaNodeDriver"),
    Provider.ALIYUN_ECS: ("libcloud.compute.drivers.ecs", "ECSDriver"),
    Provider.CLOUDSCALE: (
        "libcloud.compute.drivers.cloudscale",
        "CloudscaleNodeDriver",
    ),
    Provider.UPCLOUD: ("libcloud.compute.drivers.upcloud", "UpcloudDriver"),
    Provider.NTTCIS: ("libcloud.compute.drivers.nttcis", "NttCisNodeDriver"),
    Provider.SCALEWAY: ("libcloud.compute.drivers.scaleway", "ScalewayNodeDriver"),
    Provider.MAXIHOST: ("libcloud.compute.drivers.maxihost", "MaxihostNodeDriver"),
    Provider.GRIDSCALE: ("libcloud.compute.drivers.gridscale", "GridscaleNodeDriver"),
    Provider.KAMATERA: ("libcloud.compute.drivers.kamatera", "KamateraNodeDriver"),
    Provider.KUBEVIRT: ("libcloud.compute.drivers.kubevirt", "KubeVirtNodeDriver"),
    Provider.GIG_G8: ("libcloud.compute.drivers.gig_g8", "G8NodeDriver"),
}


def get_driver(provider):
    # type: (Union[Provider, str]) -> Type[NodeDriver]
    deprecated_constants = OLD_CONSTANT_TO_NEW_MAPPING
    return _get_provider_driver(
        drivers=DRIVERS,
        provider=provider,
        deprecated_providers=DEPRECATED_DRIVERS,
        deprecated_constants=deprecated_constants,
    )


def set_driver(provider, module, klass):
    # type: (Union[Provider, str], ModuleType, type) -> Type[NodeDriver]
    return _set_provider_driver(drivers=DRIVERS, provider=provider, module=module, klass=klass)



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
OnApp DNS Driver
"""

__all__ = ["OnAppDNSDriver"]

import json

from libcloud.dns.base import Zone, Record, DNSDriver
from libcloud.dns.types import Provider, RecordType
from libcloud.common.onapp import OnAppConnection

DEFAULT_ZONE_TTL = 1200


class OnAppDNSDriver(DNSDriver):
    type = Provider.ONAPP
    name = "OnApp"
    website = "http://onapp.com/"
    connectionCls = OnAppConnection

    RECORD_TYPE_MAP = {
        RecordType.SOA: "SOA",
        RecordType.NS: "NS",
        RecordType.A: "A",
        RecordType.AAAA: "AAAA",
        RecordType.CNAME: "CNAME",
        RecordType.MX: "MX",
        RecordType.TXT: "TXT",
        RecordType.SRV: "SRV",
    }

    def list_zones(self):
        """
        Return a list of zones.

        :return: ``list`` of :class:`Zone`
        """
        response = self.connection.request("/dns_zones.json")

        zones = self._to_zones(response.object)
        return zones

    def get_zone(self, zone_id):
        """
        Return a Zone instance.

        :param zone_id: ID of the required zone
        :type  zone_id: ``str``

        :rtype: :class:`Zone`
        """
        response = self.connection.request("/dns_zones/%s.json" % zone_id)
        zone = self._to_zone(response.object)
        return zone

    def create_zone(self, domain, type="master", ttl=None, extra=None):
        """
        Create a new zone.

        :param domain: Zone domain name (e.g. example.com)
        :type domain: ``str``

        :param type: Zone type (All zones are master by design).
        :type  type: ``str``

        :param ttl: TTL for new records. (This is not really used)
        :type  ttl: ``int``

        :param extra: Extra attributes (set auto_populate: 0 if you
        don't want to auto populate with existing DNS records). (optional)
        :type extra: ``dict``

        :rtype: :class:`Zone`

        For more info, please see:
        https://docs.onapp.com/display/52API/Add+DNS+Zone
        """
        dns_zone = {"name": domain}
        if extra is not None:
            dns_zone.update(extra)
        dns_zone_data = json.dumps({"dns_zone": dns_zone})
        response = self.connection.request(
            "/dns_zones.json",
            method="POST",
            headers={"Content-type": "application/json"},
            data=dns_zone_data,
        )
        zone = self._to_zone(response.object)
        return zone

    def delete_zone(self, zone):
        """
        Delete a zone.

        Note: This will also delete all the records belonging to this zone.

        :param zone: Zone to delete.
        :type  zone: :class:`Zone`

        :rtype: ``bool``
        """
        self.connection.request("/dns_zones/%s.json" % zone.id, method="DELETE")
        return True

    def list_records(self, zone):
        """
        Return a list of records for the provided zone.

        :param zone: Zone to list records for.
        :type zone: :class:`Zone`

        :return: ``list`` of :class:`Record`
        """
        response = self.connection.request("/dns_zones/%s/records.json" % zone.id)
        dns_records = response.object["dns_zone"]["records"]
        records = self._to_records(dns_records, zone)
        return records

    def get_record(self, zone_id, record_id):
        """
        Return a Record instance.

        :param zone_id: ID of the required zone
        :type  zone_id: ``str``

        :param record_id: ID of the required record
        :type  record_id: ``str``

        :rtype: :class:`Record`
        """
        response = self.connection.request(
            "/dns_zones/{}/records/{}.json".format(zone_id, record_id)
        )
        record = self._to_record(response.object, zone_id=zone_id)
        return record

    def create_record(self, name, zone, type, data, extra=None):
        """
        Create a new record.

        :param name: Record name without the domain name (e.g. www).
                     Note: If you want to create a record for a base domain
                     name, you should specify empty string ('') for this
                     argument.
        :type  name: ``str``

        :param zone: Zone where the requested record is created.
        :type  zone: :class:`Zone`

        :param type: DNS record type (A, AAAA, ...).
        :type  type: :class:`RecordType`

        :param data: Data for the record (depends on the record type).
        Used only for A and AAAA record types.
        :type  data: ``str``

        :param extra: Extra attributes (driver specific). (optional)
        :type extra: ``dict``

        :rtype: :class:`Record`

        For more info, please see:
        https://docs.onapp.com/display/52API/Add+DNS+Record
        """
        dns_record = self._format_record(name, type, data, extra)
        dns_record_data = json.dumps({"dns_record": dns_record})
        response = self.connection.request(
            "/dns_zones/%s/records.json" % zone.id,
            method="POST",
            headers={"Content-type": "application/json"},
            data=dns_record_data,
        )
        record = self._to_record(response.object, zone=zone)
        return record

    def update_record(self, record, name, type, data, extra=None):
        """
        Update an existing record.

        :param record: Record to update.
        :type  record: :class:`Record`

        :param name: Record name without the domain name (e.g. www).
                     Note: If you want to create a record for a base domain
                     name, you should specify empty string ('') for this
                     argument.
        :type  name: ``str``

        :param type: DNS record type (A, AAAA, ...).
        :type  type: :class:`RecordType`

        :param data: Data for the record (depends on the record type).
        Used only for A and AAAA record types.
        :type  data: ``str``

        :param extra: (optional) Extra attributes (driver specific).
        :type  extra: ``dict``

        :rtype: :class:`Record`

        For more info, please see:
        https://docs.onapp.com/display/52API/Edit+DNS+Records
        """
        zone = record.zone
        dns_record = self._format_record(name, type, data, extra)
        dns_record_data = json.dumps({"dns_record": dns_record})
        self.connection.request(
            "/dns_zones/{}/records/{}.json".format(zone.id, record.id),
            method="PUT",
            headers={"Content-type": "application/json"},
            data=dns_record_data,
        )
        record = self.get_record(zone.id, record.id)
        return record

    def delete_record(self, record):
        """
        Delete a record.

        :param record: Record to delete.
        :type  record: :class:`Record`

        :rtype: ``bool``

        For more info, please see:
        https://docs.onapp.com/display/52API/Delete+DNS+Record
        """
        zone_id = record.zone.id
        self.connection.request(
            "/dns_zones/{}/records/{}.json".format(zone_id, record.id), method="DELETE"
        )
        return True

    #
    # Helper methods
    #

    def _format_record(self, name, type, data, extra):
        if name == "":
            name = "@"
        if extra is None:
            extra = {}
        record_type = self.RECORD_TYPE_MAP[type]
        new_record = {
            "name": name,
            "ttl": extra.get("ttl", DEFAULT_ZONE_TTL),
            "type": record_type,
        }
        if type == RecordType.MX:
            additions = {
                "priority": extra.get("priority", 1),
                "hostname": extra.get("hostname"),
            }
        elif type == RecordType.SRV:
            additions = {
                "port": extra.get("port"),
                "weight": extra.get("weight", 1),
                "priority": extra.get("priority", 1),
                "hostname": extra.get("hostname"),
            }
        elif type == RecordType.A:
            additions = {"ip": data}
        elif type == RecordType.CNAME:
            additions = {"hostname": extra.get("hostname")}
        elif type == RecordType.AAAA:
            additions = {"ip": data}
        elif type == RecordType.TXT:
            additions = {"txt": extra.get("txt")}
        elif type == RecordType.NS:
            additions = {"hostname": extra.get("hostname")}

        new_record.update(additions)
        return new_record

    def _to_zones(self, data):
        zones = []
        for zone in data:
            _zone = self._to_zone(zone)
            zones.append(_zone)

        return zones

    def _to_zone(self, data):
        dns_zone = data.get("dns_zone")
        id = dns_zone.get("id")
        name = dns_zone.get("name")
        extra = {
            "user_id": dns_zone.get("user_id"),
            "cdn_reference": dns_zone.get("cdn_reference"),
            "created_at": dns_zone.get("created_at"),
            "updated_at": dns_zone.get("updated_at"),
        }

        type = "master"

        return Zone(
            id=id,
            domain=name,
            type=type,
            ttl=DEFAULT_ZONE_TTL,
            driver=self,
            extra=extra,
        )

    def _to_records(self, data, zone):
        records = []
        data = data.values()
        for data_type in data:
            for item in data_type:
                record = self._to_record(item, zone=zone)
                records.append(record)
        records.sort(key=lambda x: x.id, reverse=False)
        return records

    def _to_record(self, data, zone_id=None, zone=None):
        if not zone:  # We need zone_id or zone
            zone = self.get_zone(zone_id)
        record = data.get("dns_record")
        id = record.get("id")
        name = record.get("name")
        type = record.get("type")
        ttl = record.get("ttl", None)
        return Record(
            id=id,
            name=name,
            type=type,
            data=record,
            zone=zone,
            driver=self,
            ttl=ttl,
            extra={},
        )



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unexpected-keyword-arg

import os
import hmac
import time
import base64
import codecs
from hashlib import sha1

from libcloud.utils.py3 import ET, b, httplib, tostring, urlquote, urlencode
from libcloud.utils.xml import findtext, fixxpath
from libcloud.common.base import RawResponse, XmlResponse, ConnectionUserAndKey
from libcloud.utils.files import read_in_chunks
from libcloud.common.types import LibcloudError, InvalidCredsError, MalformedResponseError
from libcloud.storage.base import Object, Container, StorageDriver
from libcloud.storage.types import (
    ContainerError,
    ObjectDoesNotExistError,
    ObjectHashMismatchError,
    ContainerIsNotEmptyError,
    InvalidContainerNameError,
    ContainerDoesNotExistError,
)

try:
    from lxml.etree import Element, SubElement
except ImportError:
    from xml.etree.ElementTree import Element, SubElement


__all__ = [
    "OSSStorageDriver",
    "OSSMultipartUpload",
    "EXPIRATION_SECONDS",
    "CHUNK_SIZE",
    "MAX_UPLOADS_PER_RESPONSE",
]

GMT_TIME_FORMAT = "%a, %d %b %Y %H:%M:%S GMT"
EXPIRATION_SECONDS = 15 * 60

# OSS multi-part chunks must be great than 100KB except the last one
CHUNK_SIZE = 100 * 1024

# Desired number of items in each response inside a paginated request in
# ex_iterate_multipart_uploads.
MAX_UPLOADS_PER_RESPONSE = 1000


class OSSResponse(XmlResponse):
    namespace = None
    valid_response_codes = [httplib.NOT_FOUND, httplib.CONFLICT, httplib.BAD_REQUEST]

    def success(self):
        i = int(self.status)
        return 200 <= i <= 299 or i in self.valid_response_codes

    def parse_body(self):
        """
        OSSResponse body is in utf-8 encoding.
        """
        if len(self.body) == 0 and not self.parse_zero_length_body:
            return self.body

        try:
            parser = ET.XMLParser(encoding="utf-8")
            body = ET.XML(self.body.encode("utf-8"), parser=parser)
        except Exception:
            raise MalformedResponseError(
                "Failed to parse XML", body=self.body, driver=self.connection.driver
            )
        return body

    def parse_error(self):
        if self.status in [httplib.UNAUTHORIZED, httplib.FORBIDDEN]:
            raise InvalidCredsError(self.body)
        elif self.status == httplib.MOVED_PERMANENTLY:
            raise LibcloudError(
                "This bucket is located in a different " + "region. Please use the correct driver.",
                driver=OSSStorageDriver,
            )
        elif self.status == httplib.METHOD_NOT_ALLOWED:
            raise LibcloudError(
                "The method is not allowed. Status code: %d, "
                "headers: %s" % (self.status, self.headers)
            )
        raise LibcloudError(
            "Unknown error. Status code: %d, body: %s" % (self.status, self.body),
            driver=OSSStorageDriver,
        )


class OSSRawResponse(OSSResponse, RawResponse):
    pass


class OSSConnection(ConnectionUserAndKey):
    """
    Represents a single connection to the Aliyun OSS Endpoint
    """

    _domain = "aliyuncs.com"
    _default_location = "oss"
    responseCls = OSSResponse
    rawResponseCls = OSSRawResponse

    @staticmethod
    def _get_auth_signature(method, headers, params, expires, secret_key, path, vendor_prefix):
        """
        Signature = base64(hmac-sha1(AccessKeySecret,
          VERB + "\n"
          + CONTENT-MD5 + "\n"
          + CONTENT-TYPE + "\n"
          + EXPIRES + "\n"
          + CanonicalizedOSSHeaders
          + CanonicalizedResource))
        """
        special_headers = {"content-md5": "", "content-type": "", "expires": ""}
        vendor_headers = {}

        for key, value in list(headers.items()):
            key_lower = key.lower()
            if key_lower in special_headers:
                special_headers[key_lower] = value.strip()
            elif key_lower.startswith(vendor_prefix):
                vendor_headers[key_lower] = value.strip()

        if expires:
            special_headers["expires"] = str(expires)

        buf = [method]
        for _, value in sorted(special_headers.items()):
            buf.append(value)
        string_to_sign = "\n".join(buf)

        buf = []
        for key, value in sorted(vendor_headers.items()):
            buf.append("{}:{}".format(key, value))
        header_string = "\n".join(buf)

        values_to_sign = []
        for value in [string_to_sign, header_string, path]:
            if value:
                values_to_sign.append(value)

        string_to_sign = "\n".join(values_to_sign)
        b64_hmac = base64.b64encode(
            hmac.new(b(secret_key), b(string_to_sign), digestmod=sha1).digest()
        )
        return b64_hmac

    @staticmethod
    def _get_expires(params):
        """
        Get expires timeout seconds from parameters.
        """
        expires = None
        if "expires" in params:
            expires = params["expires"]
        elif "Expires" in params:
            expires = params["Expires"]
        if expires:
            try:
                return int(expires)
            except Exception:
                pass
        return int(time.time()) + EXPIRATION_SECONDS

    def add_default_params(self, params):
        expires_at = self._get_expires(params)
        expires = str(expires_at)
        params["OSSAccessKeyId"] = self.user_id
        params["Expires"] = expires
        return params

    def add_default_headers(self, headers):
        headers["Date"] = time.strftime(GMT_TIME_FORMAT, time.gmtime())
        return headers

    def pre_connect_hook(self, params, headers):
        if self._container:
            path = "/{}{}".format(self._container.name, self.action)
        else:
            path = self.action

        # pylint: disable=no-member
        params["Signature"] = self._get_auth_signature(
            method=self.method,
            headers=headers,
            params=params,
            expires=params["Expires"],
            secret_key=self.key,
            path=path,
            vendor_prefix=self.driver.http_vendor_prefix,
        )
        return params, headers

    def request(
        self,
        action,
        params=None,
        data=None,
        headers=None,
        method="GET",
        raw=False,
        container=None,
    ):
        self.host = "{}.{}".format(self._default_location, self._domain)
        self._container = container
        if container and container.name:
            if "location" in container.extra:
                self.host = "{}.{}.{}".format(
                    container.name,
                    container.extra["location"],
                    self._domain,
                )
            else:
                self.host = "{}.{}".format(container.name, self.host)
        return super().request(
            action=action,
            params=params,
            data=data,
            headers=headers,
            method=method,
            raw=raw,
        )


class OSSMultipartUpload:
    """
    Class representing an Aliyun OSS multipart upload
    """

    def __init__(self, key, id, initiated):
        """
        Class representing an Aliyun OSS multipart upload

        :param key: The object/key that was being uploaded
        :type key: ``str``

        :param id: The upload id assigned by Aliyun
        :type id: ``str``

        :param initiated: The date/time at which the upload was started
        :type created_at: ``str``
        """
        self.key = key
        self.id = id
        self.initiated = initiated

    def __repr__(self):
        return "<OSSMultipartUpload: key=%s>" % (self.key)


class OSSStorageDriver(StorageDriver):
    name = "Aliyun OSS"
    website = "http://www.aliyun.com/product/oss"
    connectionCls = OSSConnection
    hash_type = "md5"
    supports_chunked_encoding = False
    supports_multipart_upload = True
    namespace = None
    http_vendor_prefix = "x-oss-"

    def iterate_containers(self):
        response = self.connection.request("/")
        if response.status == httplib.OK:
            containers = self._to_containers(obj=response.object, xpath="Buckets/Bucket")
            return containers

        raise LibcloudError("Unexpected status code: %s" % (response.status), driver=self)

    def iterate_container_objects(self, container, prefix=None, ex_prefix=None):
        """
        Return a generator of objects for the given container.

        :param container: Container instance
        :type container: :class:`Container`

        :keyword prefix: Only return objects starting with prefix
        :type prefix: ``str``

        :keyword ex_prefix: (Deprecated.) Only return objects starting with
                            ex_prefix
        :type ex_prefix: ``str``

        :return: A generator of Object instances.
        :rtype: ``generator`` of :class:`Object`
        """
        prefix = self._normalize_prefix_argument(prefix, ex_prefix)

        params = {}

        if prefix:
            params["prefix"] = prefix

        last_key = None
        exhausted = False

        while not exhausted:
            if last_key:
                params["marker"] = last_key

            response = self.connection.request("/", params=params, container=container)

            if response.status != httplib.OK:
                raise LibcloudError("Unexpected status code: %s" % (response.status), driver=self)

            objects = self._to_objs(obj=response.object, xpath="Contents", container=container)
            is_truncated = response.object.findtext(
                fixxpath(xpath="IsTruncated", namespace=self.namespace)
            ).lower()
            exhausted = is_truncated == "false"

            last_key = None
            for obj in objects:
                last_key = obj.name
                yield obj

    def get_container(self, container_name):
        for container in self.iterate_containers():
            if container.name == container_name:
                return container
        raise ContainerDoesNotExistError(value=None, driver=self, container_name=container_name)

    def get_object(self, container_name, object_name):
        container = self.get_container(container_name=container_name)
        object_path = self._get_object_path(container, object_name)
        response = self.connection.request(object_path, method="HEAD", container=container)

        if response.status == httplib.OK:
            obj = self._headers_to_object(
                object_name=object_name, container=container, headers=response.headers
            )
            return obj

        raise ObjectDoesNotExistError(value=None, driver=self, object_name=object_name)

    def create_container(self, container_name, ex_location=None):
        """
        @inherits :class:`StorageDriver.create_container`

        :keyword ex_location: The desired location where to create container
        :type keyword: ``str``
        """
        extra = None
        if ex_location:
            root = Element("CreateBucketConfiguration")
            child = SubElement(root, "LocationConstraint")
            child.text = ex_location

            data = tostring(root)
            extra = {"location": ex_location}
        else:
            data = ""

        container = Container(name=container_name, extra=extra, driver=self)
        response = self.connection.request("/", data=data, method="PUT", container=container)

        if response.status == httplib.OK:
            return container
        elif response.status == httplib.CONFLICT:
            raise InvalidContainerNameError(
                value="Container with this name already exists. The name must "
                "be unique among all the containers in the system",
                container_name=container_name,
                driver=self,
            )
        elif response.status == httplib.BAD_REQUEST:
            raise ContainerError(
                value="Bad request when creating container: %s" % response.body,
                container_name=container_name,
                driver=self,
            )

        raise LibcloudError("Unexpected status code: %s" % (response.status), driver=self)

    def delete_container(self, container):
        # Note: All the objects in the container must be deleted first
        response = self.connection.request("/", method="DELETE", container=container)
        if response.status == httplib.NO_CONTENT:
            return True
        elif response.status == httplib.CONFLICT:
            raise ContainerIsNotEmptyError(
                value="Container must be empty before it can be deleted.",
                container_name=container.name,
                driver=self,
            )
        elif response.status == httplib.NOT_FOUND:
            raise ContainerDoesNotExistError(value=None, driver=self, container_name=container.name)

        return False

    def download_object(
        self, obj, destination_path, overwrite_existing=False, delete_on_failure=True
    ):
        obj_path = self._get_object_path(obj.container, obj.name)

        response = self.connection.request(
            obj_path, method="GET", raw=True, container=obj.container
        )

        return self._get_object(
            obj=obj,
            callback=self._save_object,
            response=response,
            callback_kwargs={
                "obj": obj,
                "response": response.response,
                "destination_path": destination_path,
                "overwrite_existing": overwrite_existing,
                "delete_on_failure": delete_on_failure,
            },
            success_status_code=httplib.OK,
        )

    def download_object_as_stream(self, obj, chunk_size=None):
        obj_path = self._get_object_path(obj.container, obj.name)
        response = self.connection.request(
            obj_path, method="GET", raw=True, container=obj.container
        )

        return self._get_object(
            obj=obj,
            callback=read_in_chunks,
            response=response,
            callback_kwargs={"iterator": response.response, "chunk_size": chunk_size},
            success_status_code=httplib.OK,
        )

    def upload_object(
        self,
        file_path,
        container,
        object_name,
        extra=None,
        verify_hash=True,
        headers=None,
    ):
        return self._put_object(
            container=container,
            object_name=object_name,
            extra=extra,
            file_path=file_path,
            verify_hash=verify_hash,
        )

    def upload_object_via_stream(self, iterator, container, object_name, extra=None, headers=None):
        method = "PUT"
        params = None

        if self.supports_multipart_upload:
            # @TODO: This needs implementing again from scratch.
            pass
        return self._put_object(
            container=container,
            object_name=object_name,
            extra=extra,
            method=method,
            query_args=params,
            stream=iterator,
            verify_hash=False,
            headers=headers,
        )

    def delete_object(self, obj):
        object_path = self._get_object_path(obj.container, obj.name)
        response = self.connection.request(object_path, method="DELETE", container=obj.container)
        if response.status == httplib.NO_CONTENT:
            return True
        elif response.status == httplib.NOT_FOUND:
            raise ObjectDoesNotExistError(value=None, driver=self, object_name=obj.name)

        return False

    def ex_iterate_multipart_uploads(
        self,
        container,
        prefix=None,
        delimiter=None,
        max_uploads=MAX_UPLOADS_PER_RESPONSE,
    ):
        """
        Extension method for listing all in-progress OSS multipart uploads.

        Each multipart upload which has not been committed or aborted is
        considered in-progress.

        :param container: The container holding the uploads
        :type container: :class:`Container`

        :keyword prefix: Print only uploads of objects with this prefix
        :type prefix: ``str``

        :keyword delimiter: The object/key names are grouped based on
            being split by this delimiter
        :type delimiter: ``str``

        :keyword max_uploads: The max upload items returned for one request
        :type max_uploads: ``int``

        :return: A generator of OSSMultipartUpload instances.
        :rtype: ``generator`` of :class:`OSSMultipartUpload`
        """

        if not self.supports_multipart_upload:
            raise LibcloudError("Feature not supported", driver=self)

        request_path = "/?uploads"
        params = {"max-uploads": max_uploads}

        if prefix:
            params["prefix"] = prefix

        if delimiter:
            params["delimiter"] = delimiter

        def finder(node, text):
            return node.findtext(fixxpath(xpath=text, namespace=self.namespace))

        while True:
            response = self.connection.request(request_path, params=params, container=container)

            if response.status != httplib.OK:
                raise LibcloudError(
                    "Error fetching multipart uploads. " "Got code: %s" % response.status,
                    driver=self,
                )

            body = response.parse_body()
            # pylint: disable=maybe-no-member
            for node in body.findall(fixxpath(xpath="Upload", namespace=self.namespace)):
                key = finder(node, "Key")
                upload_id = finder(node, "UploadId")
                initiated = finder(node, "Initiated")

                yield OSSMultipartUpload(key, upload_id, initiated)

            # Check if this is the last entry in the listing
            # pylint: disable=maybe-no-member
            is_truncated = body.findtext(fixxpath(xpath="IsTruncated", namespace=self.namespace))

            if is_truncated.lower() == "false":
                break

            # Provide params for the next request
            upload_marker = body.findtext(
                fixxpath(xpath="NextUploadIdMarker", namespace=self.namespace)
            )
            key_marker = body.findtext(fixxpath(xpath="NextKeyMarker", namespace=self.namespace))

            params["key-marker"] = key_marker
            params["upload-id-marker"] = upload_marker

    def ex_abort_all_multipart_uploads(self, container, prefix=None):
        """
        Extension method for removing all partially completed OSS multipart
        uploads.

        :param container: The container holding the uploads
        :type container: :class:`Container`

        :keyword prefix: Delete only uploads of objects with this prefix
        :type prefix: ``str``
        """

        # Iterate through the container and delete the upload ids
        for upload in self.ex_iterate_multipart_uploads(container, prefix, delimiter=None):
            object_path = self._get_object_path(container, upload.key)
            self._abort_multipart(object_path, upload.id, container=container)

    def _clean_object_name(self, name):
        name = urlquote(name)
        return name

    def _upload_object(
        self,
        object_name,
        content_type,
        request_path,
        request_method="PUT",
        headers=None,
        file_path=None,
        stream=None,
        chunked=False,
        multipart=False,
        container=None,
    ):
        """
        Helper function for setting common request headers and calling the
        passed in callback which uploads an object.
        """
        headers = headers or {}

        if file_path and not os.path.exists(file_path):
            raise OSError("File %s does not exist" % (file_path))

        if stream is not None and not hasattr(stream, "next") and not hasattr(stream, "__next__"):
            raise AttributeError("iterator object must implement next() " + "method.")

        headers["Content-Type"] = self._determine_content_type(
            content_type, object_name, file_path=file_path
        )

        if stream:
            response = self.connection.request(
                request_path,
                method=request_method,
                data=stream,
                headers=headers,
                raw=True,
                container=container,
            )
            stream_hash, stream_length = self._hash_buffered_stream(
                stream, self._get_hash_function()
            )
        else:
            with open(file_path, "rb") as file_stream:
                response = self.connection.request(
                    request_path,
                    method=request_method,
                    data=file_stream,
                    headers=headers,
                    raw=True,
                    container=container,
                )
            with open(file_path, "rb") as file_stream:
                stream_hash, stream_length = self._hash_buffered_stream(
                    file_stream, self._get_hash_function()
                )

        return {
            "response": response,
            "bytes_transferred": stream_length,
            "data_hash": stream_hash,
        }

    def _put_object(
        self,
        container,
        object_name,
        method="PUT",
        query_args=None,
        extra=None,
        file_path=None,
        stream=None,
        verify_hash=False,
        headers=None,
    ):
        """
        Create an object and upload data using the given function.
        """
        headers = headers or {}
        extra = extra or {}

        content_type = extra.get("content_type", None)
        meta_data = extra.get("meta_data", None)
        acl = extra.get("acl", None)

        if meta_data:
            for key, value in list(meta_data.items()):
                key = self.http_vendor_prefix + "meta-%s" % (key)
                headers[key] = value

        if acl:
            if acl not in ["public-read", "private", "public-read-write"]:
                raise AttributeError("invalid acl value: %s" % acl)
            headers[self.http_vendor_prefix + "object-acl"] = acl

        request_path = self._get_object_path(container, object_name)

        if query_args:
            request_path = "?".join((request_path, query_args))

        result_dict = self._upload_object(
            object_name=object_name,
            content_type=content_type,
            request_path=request_path,
            request_method=method,
            headers=headers,
            file_path=file_path,
            stream=stream,
            container=container,
        )

        response = result_dict["response"]
        bytes_transferred = result_dict["bytes_transferred"]
        headers = response.headers

        server_hash = headers["etag"].replace('"', "")

        if verify_hash and result_dict["data_hash"].upper() != server_hash.upper():
            raise ObjectHashMismatchError(
                value="MD5 hash {} checksum does not match {}".format(
                    server_hash, result_dict["data_hash"]
                ),
                object_name=object_name,
                driver=self,
            )
        elif response.status == httplib.OK:
            obj = Object(
                name=object_name,
                size=bytes_transferred,
                hash=server_hash,
                extra={"acl": acl},
                meta_data=meta_data,
                container=container,
                driver=self,
            )

            return obj
        else:
            raise LibcloudError(
                "Unexpected status code, status_code=%s" % (response.status),
                driver=self,
            )

    def _upload_multipart(
        self, response, data, iterator, container, object_name, calculate_hash=True
    ):
        """
        Callback invoked for uploading data to OSS using Aliyun's
        multipart upload mechanism

        :param response: Response object from the initial POST request
        :type response: :class:`OSSRawResponse`

        :param data: Any data from the initial POST request
        :type data: ``str``

        :param iterator: The generator for fetching the upload data
        :type iterator: ``generator``

        :param container: The container owning the object to which data is
            being uploaded
        :type container: :class:`Container`

        :param object_name: The name of the object to which we are uploading
        :type object_name: ``str``

        :keyword calculate_hash: Indicates if we must calculate the data hash
        :type calculate_hash: ``bool``

        :return: A tuple of (status, checksum, bytes transferred)
        :rtype: ``tuple``
        """

        object_path = self._get_object_path(container, object_name)

        # Get the upload id from the response xml
        response.body = response.response.read()
        body = response.parse_body()
        upload_id = body.find(fixxpath(xpath="UploadId", namespace=self.namespace)).text

        try:
            # Upload the data through the iterator
            result = self._upload_from_iterator(
                iterator, object_path, upload_id, calculate_hash, container=container
            )
            (chunks, data_hash, bytes_transferred) = result

            # Commit the chunk info and complete the upload
            etag = self._commit_multipart(object_path, upload_id, chunks, container=container)
        except Exception as e:
            # Amazon provides a mechanism for aborting an upload.
            self._abort_multipart(object_path, upload_id, container=container)
            raise e

        # Modify the response header of the first request. This is used
        # by other functions once the callback is done
        response.headers["etag"] = etag

        return (True, data_hash, bytes_transferred)

    def _upload_from_iterator(
        self, iterator, object_path, upload_id, calculate_hash=True, container=None
    ):
        """
        Uploads data from an iterator in fixed sized chunks to OSS

        :param iterator: The generator for fetching the upload data
        :type iterator: ``generator``

        :param object_path: The path of the object to which we are uploading
        :type object_name: ``str``

        :param upload_id: The upload id allocated for this multipart upload
        :type upload_id: ``str``

        :keyword calculate_hash: Indicates if we must calculate the data hash
        :type calculate_hash: ``bool``

        :keyword container: the container object to upload object to
        :type container: :class:`Container`

        :return: A tuple of (chunk info, checksum, bytes transferred)
        :rtype: ``tuple``
        """

        data_hash = None
        if calculate_hash:
            data_hash = self._get_hash_function()

        bytes_transferred = 0
        count = 1
        chunks = []
        params = {"uploadId": upload_id}

        # Read the input data in chunk sizes suitable for AWS
        for data in read_in_chunks(
            iterator, chunk_size=CHUNK_SIZE, fill_size=True, yield_empty=True
        ):
            bytes_transferred += len(data)

            if calculate_hash:
                data_hash.update(data)

            chunk_hash = self._get_hash_function()
            chunk_hash.update(data)
            chunk_hash = base64.b64encode(chunk_hash.digest()).decode("utf-8")

            # OSS will calculate hash of the uploaded data and
            # check this header.
            headers = {"Content-MD5": chunk_hash}
            params["partNumber"] = count

            request_path = "?".join((object_path, urlencode(params)))

            resp = self.connection.request(
                request_path,
                method="PUT",
                data=data,
                headers=headers,
                container=container,
            )

            if resp.status != httplib.OK:
                raise LibcloudError("Error uploading chunk", driver=self)

            server_hash = resp.headers["etag"]

            # Keep this data for a later commit
            chunks.append((count, server_hash))
            count += 1

        if calculate_hash:
            data_hash = data_hash.hexdigest()

        return (chunks, data_hash, bytes_transferred)

    def _commit_multipart(self, object_path, upload_id, chunks, container=None):
        """
        Makes a final commit of the data.

        :param object_path: Server side object path.
        :type object_path: ``str``

        :param upload_id: ID of the multipart upload.
        :type upload_id: ``str``

        :param upload_id: A list of (chunk_number, chunk_hash) tuples.
        :type upload_id: ``list``

        :keyword container: The container owning the object to which data is
            being uploaded
        :type container: :class:`Container`
        """

        root = Element("CompleteMultipartUpload")

        for count, etag in chunks:
            part = SubElement(root, "Part")
            part_no = SubElement(part, "PartNumber")
            part_no.text = str(count)

            etag_id = SubElement(part, "ETag")
            etag_id.text = str(etag)

        data = tostring(root)

        params = {"uploadId": upload_id}
        request_path = "?".join((object_path, urlencode(params)))
        response = self.connection.request(
            request_path, data=data, method="POST", container=container
        )

        if response.status != httplib.OK:
            element = response.object
            # pylint: disable=maybe-no-member
            code, message = response._parse_error_details(element=element)
            msg = "Error in multipart commit: {} ({})".format(message, code)
            raise LibcloudError(msg, driver=self)

        # Get the server's etag to be passed back to the caller
        body = response.parse_body()
        server_hash = body.find(fixxpath(xpath="ETag", namespace=self.namespace)).text
        return server_hash

    def _abort_multipart(self, object_path, upload_id, container=None):
        """
        Aborts an already initiated multipart upload

        :param object_path: Server side object path.
        :type object_path: ``str``

        :param upload_id: ID of the multipart upload.
        :type upload_id: ``str``

        :keyword container: The container owning the object to which data is
            being uploaded
        :type container: :class:`Container`
        """

        params = {"uploadId": upload_id}
        request_path = "?".join((object_path, urlencode(params)))
        resp = self.connection.request(request_path, method="DELETE", container=container)

        if resp.status != httplib.NO_CONTENT:
            raise LibcloudError(
                "Error in multipart abort. status_code=%d" % (resp.status), driver=self
            )

    def _to_containers(self, obj, xpath):
        for element in obj.findall(fixxpath(xpath=xpath, namespace=self.namespace)):
            yield self._to_container(element)

    def _to_container(self, element):
        extra = {
            "creation_date": findtext(
                element=element, xpath="CreationDate", namespace=self.namespace
            ),
            "location": findtext(element=element, xpath="Location", namespace=self.namespace),
        }

        container = Container(
            name=findtext(element=element, xpath="Name", namespace=self.namespace),
            extra=extra,
            driver=self,
        )

        return container

    def _to_objs(self, obj, xpath, container):
        return [
            self._to_obj(element, container)
            for element in obj.findall(fixxpath(xpath=xpath, namespace=self.namespace))
        ]

    def _to_obj(self, element, container):
        owner_id = findtext(element=element, xpath="Owner/ID", namespace=self.namespace)
        owner_display_name = findtext(
            element=element, xpath="Owner/DisplayName", namespace=self.namespace
        )
        meta_data = {
            "owner": {
                "id": owner_id,
                "display_name": self._safe_decode(owner_display_name),
            }
        }
        last_modified = findtext(element=element, xpath="LastModified", namespace=self.namespace)
        extra = {"last_modified": last_modified}

        name = self._safe_decode(findtext(element=element, xpath="Key", namespace=self.namespace))
        obj = Object(
            name=name,
            size=int(findtext(element=element, xpath="Size", namespace=self.namespace)),
            hash=findtext(element=element, xpath="ETag", namespace=self.namespace).replace('"', ""),
            extra=extra,
            meta_data=meta_data,
            container=container,
            driver=self,
        )

        return obj

    def _safe_decode(self, encoded):
        """
        Decode it as an escaped string and then treat the content as
        UTF-8 encoded.
        """
        try:
            if encoded:
                unescaped, _ign = codecs.escape_decode(encoded)
                return unescaped.decode("utf-8")
            return encoded
        except Exception:
            return encoded

    def _get_container_path(self, container):
        """
        Return a container path

        :param container: Container instance
        :type  container: :class:`Container`

        :return: A path for this container.
        :rtype: ``str``
        """
        return "/%s" % (container.name)

    def _get_object_path(self, container, object_name):
        """
        Return an object's path.
        Aliyun OSS api puts the container name in the host,
        so ignore container here.

        :param container: Container instance
        :type  container: :class:`Container`

        :param object_name: Object name
        :type  object_name: :class:`str`

        :return: A  path for this object.
        :rtype: ``str``
        """
        object_name_cleaned = self._clean_object_name(object_name)
        object_path = "/%s" % object_name_cleaned
        return object_path

    def _headers_to_object(self, object_name, container, headers):
        hash = headers["etag"].replace('"', "")
        extra = {"content_type": headers["content-type"], "etag": headers["etag"]}
        meta_data = {}

        if "last-modified" in headers:
            extra["last_modified"] = headers["last-modified"]

        for key, value in headers.items():
            if not key.lower().startswith(self.http_vendor_prefix + "meta-"):
                continue

            key = key.replace(self.http_vendor_prefix + "meta-", "")
            meta_data[key] = value

        obj = Object(
            name=object_name,
            size=int(headers["content-length"]),
            hash=hash,
            extra=extra,
            meta_data=meta_data,
            container=container,
            driver=self,
        )
        return obj
